from optuna.pruners.base import BasePruner  # NOQA
from optuna.pruners.median import MedianPruner  # NOQA
from optuna.pruners.nop import NopPruner  # NOQA
from optuna.pruners.percentile import PercentilePruner  # NOQA
from optuna.pruners.successive_halving import SuccessiveHalvingPruner  # NOQA



from __future__ import print_function

import optuna
from optuna import type_checking

if type_checking.TYPE_CHECKING:
    from typing import Any  # NOQA

try:
    from fastai.basic_train import Learner  # NOQA
    from fastai.callbacks import TrackerCallback
    _available = True
except ImportError as e:
    _import_error = e
    _available = False
    TrackerCallback = object


class FastAIPruningCallback(TrackerCallback):
    """FastAI callback to prune unpromising trials for fastai.

    .. note::
        This callback is for fastai<2.0, not the coming version developed in fastai/fastai_dev.

    Example:

        Add a pruning callback which monitors validation loss directly to ``Learner``.

        .. code::

            # If registering this callback in construction
            from functools import partial

            learn = Learner(
                data, model,
                callback_fns=[partial(FastAIPruningCallback, trial=trial, monitor='valid_loss')])

    Example:

        Register a pruning callback to ``learn.fit`` and ``learn.fit_one_cycle``.

        .. code::

            learn.fit(n_epochs, callbacks=[FastAIPruningCallback(learn, trial, 'valid_loss')])
            learn.fit_one_cycle(
                n_epochs, cyc_len, max_lr,
                callbacks=[FastAIPruningCallback(learn, trial, 'valid_loss')])


    Args:
        learn:
            `fastai.basic_train.Learner <https://docs.fast.ai/basic_train.html#Learner>`_.
        trial:
            A :class:`~optuna.trial.Trial` corresponding to the current
            evaluation of the objective function.
        monitor:
            An evaluation metric for pruning, e.g. ``valid_loss`` and ``Accuracy``.
            Please refer to `fastai.Callback reference
            <https://docs.fast.ai/callback.html#Callback>`_ for further
            details.
    """

    def __init__(self, learn, trial, monitor):
        # type: (Learner, optuna.trial.Trial, str) -> None

        super(FastAIPruningCallback, self).__init__(learn, monitor)

        _check_fastai_availability()

        self.trial = trial

    def on_epoch_end(self, epoch, **kwargs):
        # type: (int, Any) -> None

        value = self.get_monitor_value()
        if value is None:
            return

        # This conversion is necessary to avoid problems reported in issues.
        # - https://github.com/pfnet/optuna/issue/642
        # - https://github.com/pfnet/optuna/issue/655.
        self.trial.report(float(value), step=epoch)
        if self.trial.should_prune():
            message = 'Trial was pruned at epoch {}.'.format(epoch)
            raise optuna.structs.TrialPruned(message)


def _check_fastai_availability():
    # type: () -> None

    if not _available:
        raise ImportError(
            'fastai is not available. Please install fastai to use this feature. '
            'fastai can be installed by executing `$ pip install fastai`. '
            'For further information, please refer to the installation guide of fastai. '
            '(The actual import error is as follows: ' + str(_import_error) + ')')



from __future__ import absolute_import

import optuna

try:
    import mxnet as mx  # NOQA
    _available = True
except ImportError as e:
    _import_error = e
    _available = False


class MXNetPruningCallback(object):
    """MXNet callback to prune unpromising trials.

    Example:

        Add a pruning callback which observes validation accuracy.

        .. code::

            model.fit(train_data=X, eval_data=Y,
                      eval_end_callback=MXNetPruningCallback(trial, eval_metric='accuracy'))

    Args:
        trial:
            A :class:`~optuna.trial.Trial` corresponding to the current evaluation of the
            objective function.
        eval_metric:
            An evaluation metric name for pruning, e.g., ``cross-entropy`` and
            ``accuracy``. If using default metrics like mxnet.metrics.Accuracy, use it's
            default metric name. For custom metrics, use the metric_name provided to
            constructor. Please refer to `mxnet.metrics reference
            <https://mxnet.apache.org/api/python/metric/metric.html>`_ for further details.
    """

    def __init__(self, trial, eval_metric):
        # type: (optuna.trial.Trial, str) -> None

        _check_mxnet_availability()

        self.trial = trial
        self.eval_metric = eval_metric

    def __call__(self, param):
        # type: (mx.model.BatchEndParams,) -> None

        if param.eval_metric is not None:
            metric_names, metric_values = param.eval_metric.get()
            if type(metric_names) == list and self.eval_metric in metric_names:
                current_score = metric_values[metric_names.index(self.eval_metric)]
            elif metric_names == self.eval_metric:
                current_score = metric_values
            else:
                raise ValueError('The entry associated with the metric name "{}" '
                                 'is not found in the evaluation result list {}.'.format(
                                     self.eval_metric, str(metric_names)))
            self.trial.report(current_score, step=param.epoch)
            if self.trial.should_prune():
                message = "Trial was pruned at epoch {}.".format(param.epoch)
                raise optuna.structs.TrialPruned(message)


def _check_mxnet_availability():
    # type: () -> None

    if not _available:
        raise ImportError(
            'MXNet is not available. Please install MXNet to use this feature. '
            'MXNet can be installed by executing `$ pip install mxnet`. '
            'For further information, please refer to the installation guide of MXNet. '
            '(The actual import error is as follows: ' + str(_import_error) + ')')



import itertools
from mock import Mock  # NOQA
from mock import patch
import multiprocessing
import os
import pandas as pd
import pickle
import pytest
import threading
import time
import uuid

import optuna
from optuna.testing.storage import StorageSupplier
from optuna import type_checking

if type_checking.TYPE_CHECKING:
    from typing import Any  # NOQA
    from typing import Dict  # NOQA
    from typing import Optional  # NOQA

STORAGE_MODES = [
    'none',  # We give `None` to storage argument, so InMemoryStorage is used.
    'new',  # We always create a new sqlite DB file for each experiment.
    'common',  # We use a sqlite DB file for the whole experiments.
]

if os.getenv('INCLUDE_SLOW_TESTS') is None:
    MAX_N_TRIALS = 20
    N_JOBS_LIST = [1, 2]
    CACHE_MODES = [True]
else:
    MAX_N_TRIALS = 50
    N_JOBS_LIST = [1, 2, 10, -1]
    CACHE_MODES = [True, False]


def setup_module():
    # type: () -> None

    StorageSupplier.setup_common_tempfile()


def teardown_module():
    # type: () -> None

    StorageSupplier.teardown_common_tempfile()


def func(trial, x_max=1.0):
    # type: (optuna.trial.Trial, float) -> float

    x = trial.suggest_uniform('x', -x_max, x_max)
    y = trial.suggest_loguniform('y', 20, 30)
    z = trial.suggest_categorical('z', (-1.0, 1.0))
    return (x - 2)**2 + (y - 25)**2 + z


class Func(object):
    def __init__(self, sleep_sec=None):
        # type: (Optional[float]) -> None

        self.n_calls = 0
        self.sleep_sec = sleep_sec
        self.lock = threading.Lock()
        self.x_max = 10.0

    def __call__(self, trial):
        # type: (optuna.trial.Trial) -> float

        with self.lock:
            self.n_calls += 1
            x_max = self.x_max
            self.x_max *= 0.9

        # Sleep for testing parallelism
        if self.sleep_sec is not None:
            time.sleep(self.sleep_sec)

        value = func(trial, x_max)
        check_params(trial.params)
        return value


def check_params(params):
    # type: (Dict[str, Any]) -> None

    assert sorted(params.keys()) == ['x', 'y', 'z']


def check_value(value):
    # type: (Optional[float]) -> None

    assert isinstance(value, float)
    assert -1.0 <= value <= 12.0**2 + 5.0**2 + 1.0


def check_frozen_trial(frozen_trial):
    # type: (optuna.structs.FrozenTrial) -> None

    if frozen_trial.state == optuna.structs.TrialState.COMPLETE:
        check_params(frozen_trial.params)
        check_value(frozen_trial.value)


def check_study(study):
    # type: (optuna.Study) -> None

    for trial in study.trials:
        check_frozen_trial(trial)

    complete_trials = [t for t in study.trials if t.state == optuna.structs.TrialState.COMPLETE]
    if len(complete_trials) == 0:
        with pytest.raises(ValueError):
            study.best_params
        with pytest.raises(ValueError):
            study.best_value
        with pytest.raises(ValueError):
            study.best_trial
    else:
        check_params(study.best_params)
        check_value(study.best_value)
        check_frozen_trial(study.best_trial)


def test_optimize_trivial_in_memory_new():
    # type: () -> None

    study = optuna.create_study()
    study.optimize(func, n_trials=10)
    check_study(study)


def test_optimize_trivial_in_memory_resume():
    # type: () -> None

    study = optuna.create_study()
    study.optimize(func, n_trials=10)
    study.optimize(func, n_trials=10)
    check_study(study)


def test_optimize_trivial_rdb_resume_study():
    # type: () -> None

    study = optuna.create_study('sqlite:///:memory:')
    study.optimize(func, n_trials=10)
    check_study(study)


def test_optimize_with_direction():
    # type: () -> None

    study = optuna.create_study(direction='minimize')
    study.optimize(func, n_trials=10)
    assert study.direction == optuna.structs.StudyDirection.MINIMIZE
    check_study(study)

    study = optuna.create_study(direction='maximize')
    study.optimize(func, n_trials=10)
    assert study.direction == optuna.structs.StudyDirection.MAXIMIZE
    check_study(study)

    with pytest.raises(ValueError):
        optuna.create_study(direction='test')


@pytest.mark.parametrize(
    'n_trials, n_jobs, storage_mode, cache_mode',
    itertools.product(
        (0, 1, 2, MAX_N_TRIALS),  # n_trials
        N_JOBS_LIST,  # n_jobs
        STORAGE_MODES,  # storage_mode
        CACHE_MODES,  # cache_mode
    ))
def test_optimize_parallel(n_trials, n_jobs, storage_mode, cache_mode):
    # type: (int, int, str, bool)-> None

    f = Func()

    with StorageSupplier(storage_mode, cache_mode) as storage:
        study = optuna.create_study(storage=storage)
        study.optimize(f, n_trials=n_trials, n_jobs=n_jobs)
        assert f.n_calls == len(study.trials) == n_trials
        check_study(study)


@pytest.mark.parametrize(
    'n_trials, n_jobs, storage_mode, cache_mode',
    itertools.product(
        (0, 1, 2, MAX_N_TRIALS, None),  # n_trials
        N_JOBS_LIST,  # n_jobs
        STORAGE_MODES,  # storage_mode
        CACHE_MODES,  # cache_mode
    ))
def test_optimize_parallel_timeout(n_trials, n_jobs, storage_mode, cache_mode):
    # type: (int, int, str, bool) -> None

    sleep_sec = 0.1
    timeout_sec = 1.0
    f = Func(sleep_sec=sleep_sec)

    with StorageSupplier(storage_mode, cache_mode) as storage:
        study = optuna.create_study(storage=storage)
        study.optimize(f, n_trials=n_trials, n_jobs=n_jobs, timeout=timeout_sec)

        assert f.n_calls == len(study.trials)

        if n_trials is not None:
            assert f.n_calls <= n_trials

        # A thread can process at most (timeout_sec / sleep_sec + 1) trials.
        n_jobs_actual = n_jobs if n_jobs != -1 else multiprocessing.cpu_count()
        max_calls = (timeout_sec / sleep_sec + 1) * n_jobs_actual
        assert f.n_calls <= max_calls

        check_study(study)


@pytest.mark.parametrize('storage_mode', STORAGE_MODES)
@pytest.mark.parametrize('cache_mode', CACHE_MODES)
def test_optimize_with_catch(storage_mode, cache_mode):
    # type: (str, bool) -> None

    with StorageSupplier(storage_mode, cache_mode) as storage:
        study = optuna.create_study(storage=storage)

        def func_value_error(_):
            # type: (optuna.trial.Trial) -> float

            raise ValueError

        # Test default exceptions.
        with pytest.raises(ValueError):
            study.optimize(func_value_error, n_trials=20)
        assert len(study.trials) == 1
        assert all(trial.state == optuna.structs.TrialState.FAIL for trial in study.trials)

        # Test acceptable exception.
        study.optimize(func_value_error, n_trials=20, catch=(ValueError, ))
        assert len(study.trials) == 21
        assert all(trial.state == optuna.structs.TrialState.FAIL for trial in study.trials)

        # Test trial with unacceptable exception.
        with pytest.raises(ValueError):
            study.optimize(func_value_error, n_trials=20, catch=(ArithmeticError, ))
        assert len(study.trials) == 22
        assert all(trial.state == optuna.structs.TrialState.FAIL for trial in study.trials)


@pytest.mark.parametrize('catch', [[], [Exception], None, 1])
def test_optimize_with_catch_invalid_type(catch):
    # type: (Any) -> None

    study = optuna.create_study()

    def func_value_error(_):
        # type: (optuna.trial.Trial) -> float

        raise ValueError

    with pytest.raises(TypeError):
        study.optimize(func_value_error, n_trials=20, catch=catch)


@pytest.mark.parametrize('storage_mode', STORAGE_MODES)
@pytest.mark.parametrize('cache_mode', CACHE_MODES)
def test_study_set_and_get_user_attrs(storage_mode, cache_mode):
    # type: (str, bool) -> None

    with StorageSupplier(storage_mode, cache_mode) as storage:
        study = optuna.create_study(storage=storage)

        study.set_user_attr('dataset', 'MNIST')
        assert study.user_attrs['dataset'] == 'MNIST'


@pytest.mark.parametrize('storage_mode', STORAGE_MODES)
@pytest.mark.parametrize('cache_mode', CACHE_MODES)
def test_study_set_and_get_system_attrs(storage_mode, cache_mode):
    # type: (str, bool) -> None

    with StorageSupplier(storage_mode, cache_mode) as storage:
        study = optuna.create_study(storage=storage)

        study.set_system_attr('system_message', 'test')
        assert study.system_attrs['system_message'] == 'test'


@pytest.mark.parametrize('storage_mode', STORAGE_MODES)
@pytest.mark.parametrize('cache_mode', CACHE_MODES)
def test_trial_set_and_get_user_attrs(storage_mode, cache_mode):
    # type: (str, bool) -> None

    def f(trial):
        # type: (optuna.trial.Trial) -> float

        trial.set_user_attr('train_accuracy', 1)
        assert trial.user_attrs['train_accuracy'] == 1
        return 0.0

    with StorageSupplier(storage_mode, cache_mode) as storage:
        study = optuna.create_study(storage=storage)
        study.optimize(f, n_trials=1)
        frozen_trial = study.trials[0]
        assert frozen_trial.user_attrs['train_accuracy'] == 1


@pytest.mark.parametrize('storage_mode', STORAGE_MODES)
@pytest.mark.parametrize('cache_mode', CACHE_MODES)
def test_trial_set_and_get_system_attrs(storage_mode, cache_mode):
    # type: (str, bool) -> None

    def f(trial):
        # type: (optuna.trial.Trial) -> float

        trial.set_system_attr('system_message', 'test')
        assert trial.system_attrs['system_message'] == 'test'
        return 0.0

    with StorageSupplier(storage_mode, cache_mode) as storage:
        study = optuna.create_study(storage=storage)
        study.optimize(f, n_trials=1)
        frozen_trial = study.trials[0]
        assert frozen_trial.system_attrs['system_message'] == 'test'


@pytest.mark.parametrize('storage_mode', STORAGE_MODES)
@pytest.mark.parametrize('cache_mode', CACHE_MODES)
def test_get_all_study_summaries(storage_mode, cache_mode):
    # type: (str, bool) -> None

    with StorageSupplier(storage_mode, cache_mode) as storage:
        study = optuna.create_study(storage=storage)
        study.optimize(Func(), n_trials=5)

        summaries = optuna.get_all_study_summaries(study._storage)
        summary = [s for s in summaries if s.study_id == study.study_id][0]

        assert summary.study_name == study.study_name
        assert summary.n_trials == 5


@pytest.mark.parametrize('storage_mode', STORAGE_MODES)
@pytest.mark.parametrize('cache_mode', CACHE_MODES)
def test_get_all_study_summaries_with_no_trials(storage_mode, cache_mode):
    # type: (str, bool) -> None

    with StorageSupplier(storage_mode, cache_mode) as storage:
        study = optuna.create_study(storage=storage)

        summaries = optuna.get_all_study_summaries(study._storage)
        summary = [s for s in summaries if s.study_id == study.study_id][0]

        assert summary.study_name == study.study_name
        assert summary.n_trials == 0
        assert summary.datetime_start is None


@pytest.mark.parametrize('storage_mode', STORAGE_MODES)
@pytest.mark.parametrize('cache_mode', CACHE_MODES)
def test_run_trial(storage_mode, cache_mode):
    # type: (str, bool) -> None

    with StorageSupplier(storage_mode, cache_mode) as storage:
        study = optuna.create_study(storage=storage)

        # Test trial without exception.
        study._run_trial(func, catch=(Exception, ), gc_after_trial=True)
        check_study(study)

        # Test trial with acceptable exception.
        def func_value_error(_):
            # type: (optuna.trial.Trial) -> float

            raise ValueError

        trial = study._run_trial(func_value_error, catch=(ValueError, ), gc_after_trial=True)
        frozen_trial = study._storage.get_trial(trial._trial_id)

        expected_message = 'Setting status of trial#1 as TrialState.FAIL because of the ' \
                           'following error: ValueError()'
        assert frozen_trial.state == optuna.structs.TrialState.FAIL
        assert frozen_trial.system_attrs['fail_reason'] == expected_message

        # Test trial with unacceptable exception.
        with pytest.raises(ValueError):
            study._run_trial(func_value_error, catch=(ArithmeticError, ), gc_after_trial=True)

        # Test trial with invalid objective value: None
        def func_none(_):
            # type: (optuna.trial.Trial) -> float

            return None  # type: ignore

        trial = study._run_trial(func_none, catch=(Exception, ), gc_after_trial=True)
        frozen_trial = study._storage.get_trial(trial._trial_id)

        expected_message = 'Setting status of trial#3 as TrialState.FAIL because the returned ' \
                           'value from the objective function cannot be casted to float. ' \
                           'Returned value is: None'
        assert frozen_trial.state == optuna.structs.TrialState.FAIL
        assert frozen_trial.system_attrs['fail_reason'] == expected_message

        # Test trial with invalid objective value: nan
        def func_nan(_):
            # type: (optuna.trial.Trial) -> float

            return float('nan')

        trial = study._run_trial(func_nan, catch=(Exception, ), gc_after_trial=True)
        frozen_trial = study._storage.get_trial(trial._trial_id)

        expected_message = 'Setting status of trial#4 as TrialState.FAIL because the objective ' \
                           'function returned nan.'
        assert frozen_trial.state == optuna.structs.TrialState.FAIL
        assert frozen_trial.system_attrs['fail_reason'] == expected_message


def test_study_pickle():
    # type: () -> None

    study_1 = optuna.create_study()
    study_1.optimize(func, n_trials=10)
    check_study(study_1)
    assert len(study_1.trials) == 10
    dumped_bytes = pickle.dumps(study_1)

    study_2 = pickle.loads(dumped_bytes)
    check_study(study_2)
    assert len(study_2.trials) == 10

    study_2.optimize(func, n_trials=10)
    check_study(study_2)
    assert len(study_2.trials) == 20


def test_study_trials_dataframe_with_no_trials():
    # type: () -> None

    study_with_no_trials = optuna.create_study()
    trials_df = study_with_no_trials.trials_dataframe()
    assert trials_df.empty


@pytest.mark.parametrize('storage_mode', STORAGE_MODES)
@pytest.mark.parametrize('cache_mode', CACHE_MODES)
@pytest.mark.parametrize('include_internal_fields', [True, False])
def test_trials_dataframe(storage_mode, cache_mode, include_internal_fields):
    # type: (str, bool, bool) -> None

    def f(trial):
        # type: (optuna.trial.Trial) -> float

        x = trial.suggest_int('x', 1, 1)
        y = trial.suggest_categorical('y', (2.5, ))
        trial.set_user_attr('train_loss', 3)
        return x + y  # 3.5

    with StorageSupplier(storage_mode, cache_mode) as storage:
        study = optuna.create_study(storage=storage)
        study.optimize(f, n_trials=3)
        df = study.trials_dataframe(include_internal_fields=include_internal_fields)
        # Change index to access rows via trial number.
        df.set_index(('number', ''), inplace=True, drop=False)
        assert len(df) == 3
        # TODO(Yanase): Remove number from system_attrs after adding TrialModel.number.
        # non-nested: 5, params: 2, user_attrs: 1, system_attrs: 1 and 9 in total.
        if include_internal_fields:
            # distributions:2, trial_id: 1
            assert len(df.columns) == 7 + 5
        else:
            assert len(df.columns) == 9

        for i in range(3):
            assert df.number[i] == i
            assert df.state[i] == optuna.structs.TrialState.COMPLETE
            assert df.value[i] == 3.5
            assert isinstance(df.datetime_start[i], pd.Timestamp)
            assert isinstance(df.datetime_complete[i], pd.Timestamp)
            assert df.params.x[i] == 1
            assert df.params.y[i] == 2.5
            assert df.user_attrs.train_loss[i] == 3
            assert df.system_attrs._number[i] == i
            if include_internal_fields:
                assert ('distributions', 'x') in df.columns
                assert ('distributions', 'y') in df.columns
                assert ('trial_id', '') in df.columns  # trial_id depends on other tests.
                assert ('distributions', 'x') in df.columns
                assert ('distributions', 'y') in df.columns


@pytest.mark.parametrize('storage_mode', STORAGE_MODES)
@pytest.mark.parametrize('cache_mode', CACHE_MODES)
def test_trials_dataframe_with_failure(storage_mode, cache_mode):
    # type: (str, bool) -> None

    def f(trial):
        # type: (optuna.trial.Trial) -> float

        x = trial.suggest_int('x', 1, 1)
        y = trial.suggest_categorical('y', (2.5, ))
        trial.set_user_attr('train_loss', 3)
        raise ValueError()
        return x + y  # 3.5

    with StorageSupplier(storage_mode, cache_mode) as storage:
        study = optuna.create_study(storage=storage)
        study.optimize(f, n_trials=3, catch=(ValueError,))
        df = study.trials_dataframe()
        # Change index to access rows via trial number.
        df.set_index(('number', ''), inplace=True, drop=False)
        assert len(df) == 3
        # TODO(Yanase): Remove number from system_attrs after adding TrialModel.number.
        # non-nested: 5, params: 2, user_attrs: 1 system_attrs: 2
        assert len(df.columns) == 10
        for i in range(3):
            assert df.number[i] == i
            assert df.state[i] == optuna.structs.TrialState.FAIL
            assert df.value[i] is None
            assert isinstance(df.datetime_start[i], pd.Timestamp)
            assert isinstance(df.datetime_complete[i], pd.Timestamp)
            assert df.params.x[i] == 1
            assert df.params.y[i] == 2.5
            assert df.user_attrs.train_loss[i] == 3
            assert df.system_attrs._number[i] == i
            assert ('system_attrs', 'fail_reason') in df.columns


@pytest.mark.parametrize('storage_mode', STORAGE_MODES)
@pytest.mark.parametrize('cache_mode', CACHE_MODES)
def test_create_study(storage_mode, cache_mode):
    # type: (str, bool) -> None

    with StorageSupplier(storage_mode, cache_mode) as storage:
        # Test creating a new study.
        study = optuna.create_study(storage=storage, load_if_exists=False)

        # Test `load_if_exists=True` with existing study.
        optuna.create_study(study_name=study.study_name, storage=storage, load_if_exists=True)

        if isinstance(study._storage, optuna.storages.InMemoryStorage):
            # `InMemoryStorage` does not share study's namespace (i.e., no name conflicts occur).
            optuna.create_study(study_name=study.study_name, storage=storage, load_if_exists=False)
        else:
            # Test `load_if_exists=False` with existing study.
            with pytest.raises(optuna.structs.DuplicatedStudyError):
                optuna.create_study(study_name=study.study_name,
                                    storage=storage,
                                    load_if_exists=False)


@pytest.mark.parametrize('storage_mode', STORAGE_MODES)
@pytest.mark.parametrize('cache_mode', CACHE_MODES)
def test_load_study(storage_mode, cache_mode):
    # type: (str, bool) -> None

    with StorageSupplier(storage_mode, cache_mode) as storage:
        if storage is None:
            # `InMemoryStorage` can not be used with `load_study` function.
            return

        study_name = str(uuid.uuid4())

        with pytest.raises(ValueError):
            # Test loading an unexisting study.
            optuna.study.load_study(study_name=study_name, storage=storage)

        # Create a new study.
        created_study = optuna.study.create_study(study_name=study_name, storage=storage)

        # Test loading an existing study.
        loaded_study = optuna.study.load_study(study_name=study_name, storage=storage)
        assert created_study.study_id == loaded_study.study_id


@pytest.mark.parametrize('storage_mode', STORAGE_MODES)
@pytest.mark.parametrize('cache_mode', CACHE_MODES)
def test_delete_study(storage_mode, cache_mode):
    # type: (str, bool) -> None

    with StorageSupplier(storage_mode, cache_mode) as storage:
        # Get storage object because delete_study does not accept None.
        storage = optuna.storages.get_storage(storage=storage)
        assert storage is not None

        # Test deleting a non-existing study.
        with pytest.raises(ValueError):
            optuna.delete_study("invalid-study-name", storage)

        # Test deleting an existing study.
        study = optuna.create_study(storage=storage, load_if_exists=False)
        optuna.delete_study(study.study_name, storage)

        # Test failed to delete the study which is already deleted.
        if not isinstance(study._storage, optuna.storages.InMemoryStorage):
            # Skip `InMemoryStorage` because it just internally initializes trials and so on.
            with pytest.raises(ValueError):
                optuna.delete_study(study.study_name, storage)


def test_nested_optimization():
    # type: () -> None

    def objective(trial):
        # type: (optuna.trial.Trial) -> float

        with pytest.raises(RuntimeError):
            trial.study.optimize(lambda _: 0.0, n_trials=1)

        return 1.0

    study = optuna.create_study()
    study.optimize(objective, n_trials=10, catch=())


@pytest.mark.parametrize('storage_mode', STORAGE_MODES)
def test_append_trial(storage_mode):
    # type: (str) -> None

    with StorageSupplier(storage_mode) as storage:
        study = optuna.create_study(storage=storage)
        assert len(study.trials) == 0

        study._append_trial(value=0.8)
        assert len(study.trials) == 1
        assert study.best_value == 0.8


def test_storage_property():
    # type: () -> None

    study = optuna.create_study()
    assert study.storage == study._storage


@patch('optuna.study.gc.collect')
def test_optimize_with_gc(collect_mock):
    # type: (Mock) -> None

    study = optuna.create_study()
    study.optimize(func, n_trials=10, gc_after_trial=True)
    check_study(study)
    assert collect_mock.call_count == 10


@patch('optuna.study.gc.collect')
def test_optimize_without_gc(collect_mock):
    # type: (Mock) -> None

    study = optuna.create_study()
    study.optimize(func, n_trials=10, gc_after_trial=False)
    check_study(study)
    assert collect_mock.call_count == 0


@pytest.mark.parametrize('n_jobs', [1, 4])
def test_callbacks(n_jobs):
    # type: (int) -> None

    study = optuna.create_study()

    def objective(trial):
        # type: (optuna.trial.Trial) -> float

        return trial.suggest_int('x', 1, 1)

    # Empty callback list.
    study.optimize(objective, callbacks=[], n_trials=10, n_jobs=n_jobs)

    # A callback.
    values = []
    callbacks = [lambda study, trial: values.append(trial.value)]
    study.optimize(objective, callbacks=callbacks, n_trials=10, n_jobs=n_jobs)
    assert values == [1] * 10

    # Two callbacks.
    values = []
    params = []
    callbacks = [
        lambda study, trial: values.append(trial.value),
        lambda study, trial: params.append(trial.params)
    ]
    study.optimize(objective, callbacks=callbacks, n_trials=10, n_jobs=n_jobs)
    assert values == [1] * 10
    assert params == [{'x': 1}] * 10

    # If a trial is failed with an exception and the exception is caught by the study,
    # callbacks are invoked.
    states = []
    callbacks = [lambda study, trial: states.append(trial.state)]
    study.optimize(
        lambda t: 1/0, callbacks=callbacks, n_trials=10, n_jobs=n_jobs,
        catch=(ZeroDivisionError,))
    assert states == [optuna.structs.TrialState.FAIL] * 10

    # NOTE: Because `Study.optimize` blocks forever if `n_jobs` is more than `1` and
    #       an uncaught exception is raised during an optimization,
    #       we test the following scenario only when `n_jobs==1`.
    #       For the details of this problem, please see https://github.com/pfnet/optuna/issues/538.
    #
    # TODO(ohta): Fix `Study.optimize`

    if n_jobs == 1:
        # If a trial is failed with an exception and the exception isn't caught by the study,
        # callbacks aren't invoked.
        states = []
        callbacks = [lambda study, trial: states.append(trial.state)]
        with pytest.raises(ZeroDivisionError):
            study.optimize(lambda t: 1/0, callbacks=callbacks,
                           n_trials=10, n_jobs=n_jobs, catch=())
        assert states == []



from __future__ import absolute_import

import optuna

if optuna.type_checking.TYPE_CHECKING:
    from typing import Dict  # NOQA
    from typing import Optional  # NOQA

try:
    from pytorch_lightning.callbacks import EarlyStopping
    _available = True
except (ImportError, SyntaxError) as e:
    # SyntaxError is raised with Python versions below 3.6 since PyTorch Lightning does not
    # support them.
    _import_error = e
    # PyTorchLightningPruningCallback is disabled because PyTorch Lightning is not available.
    _available = False
    EarlyStopping = object


class PyTorchLightningPruningCallback(EarlyStopping):
    """PyTorch Lightning callback to prune unpromising trials.

    Example:

        Add a pruning callback which observes validation accuracy.

        .. code::

            trainer.pytorch_lightning.Trainer(
                early_stop_callback=PyTorchLightningPruningCallback(trial, monitor='avg_val_acc'))

    Args:
        trial:
            A :class:`~optuna.trial.Trial` corresponding to the current evaluation of the
            objective function.
        monitor:
            An evaluation metric for pruning, e.g., ``val_loss`` or
            ``val_acc``. The metrics are obtained from the returned dictionaries from e.g.
            ``pytorch_lightning.LightningModule.training_step`` or
            ``pytorch_lightning.LightningModule.validation_end`` and the names thus depend on
            how this dictionary is formatted.
    """

    def __init__(self, trial, monitor):
        # type: (optuna.trial.Trial, str) -> None

        super(PyTorchLightningPruningCallback, self).__init__()

        _check_pytorch_lightning_availability()

        self.trial = trial
        self.monitor = monitor

    def on_epoch_end(self, epoch, logs=None):
        # type: (int, Optional[Dict[str, float]]) -> None

        logs = logs or {}
        current_score = logs.get(self.monitor)
        if current_score is None:
            return
        self.trial.report(current_score, step=epoch)
        if self.trial.should_prune():
            message = "Trial was pruned at epoch {}.".format(epoch)
            raise optuna.structs.TrialPruned(message)


def _check_pytorch_lightning_availability():
    # type: () -> None

    if not _available:
        raise ImportError(
            'PyTorch Lightning is not available. Please install PyTorch Lightning to use this '
            'feature. PyTorch Lightning can be installed by executing `$ pip install '
            'pytorch-lightning`. For further information, please refer to the installation guide '
            'of PyTorch Lightining. (The actual import error is as follows: '
            + str(_import_error) + ')')



from datetime import datetime
import enum
from typing import Any
from typing import Dict
from typing import NamedTuple
from typing import Optional

from optuna.distributions import BaseDistribution


class TrialState(enum.Enum):
    """State of a :class:`~optuna.trial.Trial`.

    Attributes:
        RUNNING:
            The :class:`~optuna.trial.Trial` is running.
        COMPLETE:
            The :class:`~optuna.trial.Trial` has been finished without any error.
        PRUNED:
            The :class:`~optuna.trial.Trial` has been pruned with :class:`TrialPruned`.
        FAIL:
            The :class:`~optuna.trial.Trial` has failed due to an uncaught error.
    """

    RUNNING = 0
    COMPLETE = 1
    PRUNED = 2
    FAIL = 3

    def is_finished(self):
        # type: () -> bool

        return self != TrialState.RUNNING


class StudyDirection(enum.Enum):
    """Direction of a :class:`~optuna.study.Study`.

    Attributes:
        NOT_SET:
            Direction has not been set.
        MINIMIZE:
            :class:`~optuna.study.Study` minimizes the objective function.
        MAXIMIZE:
            :class:`~optuna.study.Study` maximizes the objective function.
    """

    NOT_SET = 0
    MINIMIZE = 1
    MAXIMIZE = 2


class FrozenTrial(
        NamedTuple('_BaseFrozenTrial', [
            ('number', int),
            ('state', TrialState),
            ('value', Optional[float]),
            ('datetime_start', Optional[datetime]),
            ('datetime_complete', Optional[datetime]),
            ('params', Dict[str, Any]),
            ('distributions', Dict[str, BaseDistribution]),
            ('user_attrs', Dict[str, Any]),
            ('system_attrs', Dict[str, Any]),
            ('intermediate_values', Dict[int, float]),
            ('trial_id', int),
        ])):
    """Status and results of a :class:`~optuna.trial.Trial`.

    Attributes:
        number:
            Unique and consecutive number of :class:`~optuna.trial.Trial` for each
            :class:`~optuna.study.Study`. Note that this field uses zero-based numbering.
        state:
            :class:`TrialState` of the :class:`~optuna.trial.Trial`.
        value:
            Objective value of the :class:`~optuna.trial.Trial`.
        datetime_start:
            Datetime where the :class:`~optuna.trial.Trial` started.
        datetime_complete:
            Datetime where the :class:`~optuna.trial.Trial` finished.
        params:
            Dictionary that contains suggested parameters.
        distributions:
            Dictionary that contains the distributions of :attr:`params`.
        user_attrs:
            Dictionary that contains the attributes of the :class:`~optuna.trial.Trial` set with
            :func:`optuna.trial.Trial.set_user_attr`.
        system_attrs:
            Dictionary that contains the attributes of the :class:`~optuna.trial.Trial` internally
            set by Optuna.
        intermediate_values:
            Intermediate objective values set with :func:`optuna.trial.Trial.report`.
        trial_id:
            Optuna's internal identifier of the :class:`~optuna.trial.Trial`. Note that this field
            is not supposed to be used by library users. Instead, please use :attr:`number` and
            :class:`~optuna.study.Study.study_id` to identify a :class:`~optuna.trial.Trial`.
    """

    internal_fields = ['distributions', 'trial_id']

    def _validate(self):
        # type: () -> None

        if self.datetime_start is None:
            raise ValueError('`datetime_start` is supposed to be set.')

        if self.state.is_finished():
            if self.datetime_complete is None:
                raise ValueError('`datetime_complete` is supposed to be set for a finished trial.')
        else:
            if self.datetime_complete is not None:
                raise ValueError(
                    '`datetime_complete` is supposed to not be set for a finished trial.')

        if self.state == TrialState.COMPLETE and self.value is None:
            raise ValueError('`value` is supposed to be set for a complete trial.')

        if set(self.params.keys()) != set(self.distributions.keys()):
            raise ValueError('Inconsistent parameters {} and distributions {}.'.format(
                set(self.params.keys()), set(self.distributions.keys())))

        for param_name, param_value in self.params.items():
            distribution = self.distributions[param_name]

            param_value_in_internal_repr = distribution.to_internal_repr(param_value)
            if not distribution._contains(param_value_in_internal_repr):
                raise ValueError(
                    "The value {} of parameter '{}' isn't contained in the distribution {}.".
                    format(param_value, param_name, distribution))

    @property
    def last_step(self):
        # type: () -> Optional[int]

        if len(self.intermediate_values) == 0:
            return None
        else:
            return max(self.intermediate_values.keys())


class StudySummary(
        NamedTuple('StudySummary', [('study_id', int), ('study_name', str),
                                    ('direction', StudyDirection),
                                    ('best_trial', Optional[FrozenTrial]),
                                    ('user_attrs', Dict[str, Any]),
                                    ('system_attrs', Dict[str, Any]), ('n_trials', int),
                                    ('datetime_start', Optional[datetime])])):
    """Basic attributes and aggregated results of a :class:`~optuna.study.Study`.

    See also :func:`optuna.study.get_all_study_summaries`.

    Attributes:
        study_id:
            Identifier of the :class:`~optuna.study.Study`.
        study_name:
            Name of the :class:`~optuna.study.Study`.
        direction:
            :class:`StudyDirection` of the :class:`~optuna.study.Study`.
        best_trial:
            :class:`FrozenTrial` with best objective value in the :class:`~optuna.study.Study`.
        user_attrs:
            Dictionary that contains the attributes of the :class:`~optuna.study.Study` set with
            :func:`optuna.study.Study.set_user_attr`.
        system_attrs:
            Dictionary that contains the attributes of the :class:`~optuna.study.Study` internally
            set by Optuna.
        n_trials:
            The number of trials ran in the :class:`~optuna.study.Study`.
        datetime_start:
            Datetime where the :class:`~optuna.study.Study` started.
    """


class OptunaError(Exception):
    """Base class for Optuna specific errors."""

    pass


class TrialPruned(OptunaError):
    """Exception for pruned trials.

    This error tells a trainer that the current :class:`~optuna.trial.Trial` was pruned. It is
    supposed to be raised after :func:`optuna.trial.Trial.should_prune` as shown in the following
    example.

    Example:

        .. code::

            >>> def objective(trial):
            >>>     ...
            >>>     for step in range(n_train_iter):
            >>>         ...
            >>>         if trial.should_prune():
            >>>             raise TrailPruned()
    """

    pass


class CLIUsageError(OptunaError):
    """Exception for CLI.

    CLI raises this exception when it receives invalid configuration.
    """

    pass


class StorageInternalError(OptunaError):
    """Exception for storage operation.

    This error is raised when an operation failed in backend DB of storage.
    """

    pass


class DuplicatedStudyError(OptunaError):
    """Exception for a duplicated study name.

    This error is raised when a specified study name already exists in the storage.
    """

    pass



from optuna.distributions import BaseDistribution
from optuna import type_checking

if type_checking.TYPE_CHECKING:
    from typing import Dict  # NOQA


class UnsupportedDistribution(BaseDistribution):

    def single(self):
        # type: () -> bool

        return False

    def _contains(self, param_value_in_internal_repr):
        # type: (float) -> bool

        return True

    def _asdict(self):
        # type: () -> Dict

        return {}



import tempfile

import optuna
import optuna.trial


def _create_some_study():
    # type: () -> optuna.Study

    def f(trial):
        # type: (optuna.trial.Trial) -> float

        x = trial.suggest_uniform('x', -10, 10)
        y = trial.suggest_loguniform('y', 10, 20)
        z = trial.suggest_categorical('z', (10, 20.5, 30))

        return x**2 + y**2 + z

    study = optuna.create_study()
    study.optimize(f, n_trials=100)
    return study


def test_write():
    # type: () -> None

    study = _create_some_study()

    with tempfile.NamedTemporaryFile('r') as tf:
        optuna.dashboard.write(study, tf.name)

        html = tf.read()
        assert '<body>' in html
        assert 'bokeh' in html



import warnings

import optuna
from optuna import logging
from optuna.samplers.base import BaseSampler  # NOQA
from optuna.samplers.random import RandomSampler  # NOQA
from optuna.samplers.tpe import TPESampler  # NOQA

if optuna.type_checking.TYPE_CHECKING:
    from typing import Dict  # NOQA

    from optuna.distributions import BaseDistribution  # NOQA
    from optuna.study import BaseStudy  # NOQA


def intersection_search_space(study):
    # type: (BaseStudy) -> Dict[str, BaseDistribution]
    """Return the intersection search space of the :class:`~optuna.study.BaseStudy`.

    Intersection search space contains the intersection of parameter distributions that have been
    suggested in the completed trials of the study so far.
    If there are multiple parameters that have the same name but different distributions,
    neither is included in the resulting search space
    (i.e., the parameters with dynamic value ranges are excluded).

    Returns:
        A dictionary containing the parameter names and parameter's distributions.
    """

    search_space = None
    for trial in study.trials:
        if trial.state != optuna.structs.TrialState.COMPLETE:
            continue

        if search_space is None:
            search_space = trial.distributions
            continue

        delete_list = []
        for param_name, param_distribution in search_space.items():
            if param_name not in trial.distributions:
                delete_list.append(param_name)
            elif trial.distributions[param_name] != param_distribution:
                delete_list.append(param_name)

        for param_name in delete_list:
            del search_space[param_name]

    return search_space or {}


def product_search_space(study):
    # type: (BaseStudy) -> Dict[str, BaseDistribution]
    """Return the product search space of the :class:`~optuna.study.BaseStudy`.

    .. deprecated:: 0.14.0
        Please use :func:`~optuna.samplers.intersection_search_space` instead.
    """

    warnings.warn(
        '`product_search_space` function is deprecated. '
        'Please use `intersection_search_space` function instead.', DeprecationWarning)

    logger = logging.get_logger(__name__)
    logger.warning('`product_search_space` function is deprecated. '
                   'Please use `intersection_search_space` function instead.')

    return intersection_search_space(study)



import contextlib
import copy
import time

import lightgbm as lgb
import numpy as np
import tqdm

import optuna
from optuna.integration.lightgbm_tuner.alias import _handling_alias_parameters
from optuna import type_checking


if type_checking.TYPE_CHECKING:
    from typing import Any  # NOQA
    from typing import Callable  # NOQA
    from typing import Dict  # NOQA
    from typing import Generator  # NOQA
    from typing import List  # NOQA
    from typing import Optional  # NOQA
    from typing import Tuple  # NOQA
    from typing import Union  # NOQA

    from optuna.distributions import BaseDistribution  # NOQA
    from optuna.structs import FrozenTrial  # NOQA
    from optuna.study import Study  # NOQA
    from optuna.trial import Trial  # NOQA

    VALID_SET_TYPE = Union[List[lgb.Dataset], Tuple[lgb.Dataset, ...], lgb.Dataset]


# EPS is used to ensure that a sampled parameter value is in pre-defined value range.
EPS = 1e-12


class _GridSamplerUniform1D(optuna.samplers.BaseSampler):

    def __init__(self, param_name, param_values):
        # type: (str, Any) -> None

        self.param_name = param_name
        self.param_values = tuple(param_values)
        self.value_idx = 0

    def sample_relative(self, study, trial, search_space):
        # type: (Study, FrozenTrial, Dict[str, BaseDistribution]) -> Dict[str, float]

        # todo (g-votte): Take care of distributed optimization.
        assert self.value_idx < len(self.param_values)
        param_value = self.param_values[self.value_idx]
        self.value_idx += 1
        return {self.param_name: param_value}

    def sample_independent(self, study, trial, param_name, param_distribution):
        # type: (Study, FrozenTrial, str, BaseDistribution) -> None

        raise ValueError(
            'Suggest method is called for an invalid parameter: {}.'.format(param_name))

    def infer_relative_search_space(self, study, trial):
        # type: (Study, FrozenTrial) -> Dict[str, BaseDistribution]

        distribution = optuna.distributions.UniformDistribution(-float('inf'), float('inf'))
        return {self.param_name: distribution}


class _TimeKeeper(object):
    def __init__(self):
        # type: () -> None

        self.time = time.time()

    def elapsed_secs(self):
        # type: () -> float

        return time.time() - self.time


@contextlib.contextmanager
def _timer():
    # type: () -> Generator[_TimeKeeper, None, None]

    timekeeper = _TimeKeeper()
    yield timekeeper


class BaseTuner(object):
    def __init__(
            self,
            lgbm_params=None,
            lgbm_kwargs=None
    ):
        # type: (Dict[str, Any], Dict[str,Any]) -> None

        self.lgbm_params = lgbm_params or {}
        self.lgbm_kwargs = lgbm_kwargs or {}

    def _get_booster_best_score(self, booster):
        # type: (lgb.Booster) -> float

        metric = self.lgbm_params.get('metric', 'binary_logloss')

        # todo (smly): This implementation is different logic from the LightGBM's python bindings.
        if type(metric) is str:
            pass
        elif type(metric) is list:
            metric = metric[-1]
        elif type(metric) is set:
            metric = list(metric)[-1]
        else:
            raise NotImplementedError
        valid_sets = self.lgbm_kwargs.get('valid_sets')  # type: Optional[VALID_SET_TYPE]

        if self.lgbm_kwargs.get('valid_names') is not None:
            if type(self.lgbm_kwargs['valid_names']) is str:
                valid_name = self.lgbm_kwargs['valid_names']
            elif type(self.lgbm_kwargs['valid_names']) in [list, tuple]:
                valid_name = self.lgbm_kwargs['valid_names'][-1]
            else:
                raise NotImplementedError

        elif type(valid_sets) is lgb.Dataset:
            valid_name = 'valid_0'

        elif isinstance(valid_sets, (list, tuple)) and len(valid_sets) > 0:
            valid_set_idx = len(valid_sets) - 1
            valid_name = 'valid_{}'.format(valid_set_idx)

        else:
            raise NotImplementedError

        val_score = booster.best_score[valid_name][metric]
        return val_score

    def higher_is_better(self):
        # type: () -> bool

        metric_name = self.lgbm_params.get('metric', 'binary_logloss')
        return metric_name.startswith(('auc', 'ndcg@', 'map@', 'accuracy'))

    def compare_validation_metrics(self, val_score, best_score):
        # type: (float, float) -> bool

        if self.higher_is_better():
            return val_score > best_score
        else:
            return val_score < best_score


class OptunaObjective(BaseTuner):
    """Objective for hyperparameter-tuning with Optuna."""

    def __init__(
            self,
            target_param_names,  # type: List[str]
            lgbm_params,  # type: Dict[str, Any]
            train_set,  # type: lgb.Dataset
            lgbm_kwargs,  # type: Dict[str, Any]
            best_score,  # type: float
            pbar=None,  # type: Optional[tqdm.tqdm]
    ):

        self.target_param_names = target_param_names
        self.pbar = pbar
        self.lgbm_params = lgbm_params
        self.lgbm_kwargs = lgbm_kwargs
        self.train_set = train_set

        self.report = []  # type: List[Dict[str, Any]]
        self.trial_count = 0
        self.best_score = best_score
        self.best_booster = None
        self.action = 'tune_' + '_and_'.join(self.target_param_names)

        self._check_target_names_supported()

    def _check_target_names_supported(self):
        # type: () -> None

        supported_param_names = [
            'lambda_l1',
            'lambda_l2',
            'num_leaves',
            'feature_fraction',
            'bagging_fraction',
            'bagging_freq',
            'min_child_samples',
        ]
        for target_param_name in self.target_param_names:
            if target_param_name not in supported_param_names:
                raise NotImplementedError("Parameter `{}` is not supported for tunning.")

    def __call__(self, trial):
        # type: (Trial) -> float

        pbar_fmt = "{}, val_score: {:.6f}"

        if self.pbar is not None:
            self.pbar.set_description(pbar_fmt.format(self.action, self.best_score))

        if 'lambda_l1' in self.target_param_names:
            self.lgbm_params['lambda_l1'] = trial.suggest_loguniform('lambda_l1', 1e-8, 10.0)
        if 'lambda_l2' in self.target_param_names:
            self.lgbm_params['lambda_l2'] = trial.suggest_loguniform('lambda_l2', 1e-8, 10.0)
        if 'num_leaves' in self.target_param_names:
            max_depth = self.lgbm_params.get('max_depth', 8)
            self.lgbm_params['num_leaves'] = trial.suggest_int(
                'num_leaves', 2, 2 ** max_depth)
        if 'feature_fraction' in self.target_param_names:
            # `_GridSamplerUniform1D` is used for sampling feature_fraction value.
            # The value 1.0 for the hyperparameter is always sampled.
            param_value = min(trial.suggest_uniform('feature_fraction', 0.4, 1.0 + EPS), 1.0)
            self.lgbm_params['feature_fraction'] = param_value
        if 'bagging_fraction' in self.target_param_names:
            # `TPESampler` is used for sampling bagging_fraction value.
            # The value 1.0 for the hyperparameter might by sampled.
            param_value = min(trial.suggest_uniform('bagging_fraction', 0.4, 1.0 + EPS), 1.0)
            self.lgbm_params['bagging_fraction'] = param_value
        if 'bagging_freq' in self.target_param_names:
            self.lgbm_params['bagging_freq'] = trial.suggest_int('bagging_freq', 1, 7)
        if 'min_child_samples' in self.target_param_names:
            # `_GridSamplerUniform1D` is used for sampling min_child_samples value.
            # The value 1.0 for the hyperparameter is always sampled.
            param_value = int(trial.suggest_uniform('min_child_samples', 5, 100 + EPS))
            self.lgbm_params['min_child_samples'] = param_value

        with _timer() as t:
            booster = lgb.train(self.lgbm_params, self.train_set, **self.lgbm_kwargs)

        val_score = self._get_booster_best_score(booster)
        elapsed_secs = t.elapsed_secs()
        average_iteration_time = elapsed_secs / booster.current_iteration()
        if self.compare_validation_metrics(val_score, self.best_score):
            self.best_score = val_score
            self.best_booster = booster

        if self.pbar is not None:
            self.pbar.set_description(pbar_fmt.format(self.action, self.best_score))
            self.pbar.update(1)

        self.report.append(dict(
            action=self.action,
            trial=self.trial_count,
            value=str(trial.params),
            val_score=val_score,
            elapsed_secs=elapsed_secs,
            average_iteration_time=average_iteration_time))

        self.trial_count += 1

        return val_score


class LightGBMTuner(BaseTuner):
    """Hyperparameter-tuning with Optuna for LightGBM."""

    def __init__(
            self,
            params,  # type: Dict[str, Any]
            train_set,  # type: lgb.Dataset
            num_boost_round=1000,  # type: int
            valid_sets=None,  # type: Optional[VALID_SET_TYPE]
            valid_names=None,  # type: Optional[Any]
            fobj=None,  # type: Optional[Callable[..., Any]]
            feval=None,  # type: Optional[Callable[..., Any]]
            feature_name='auto',  # type: str
            categorical_feature='auto',  # type: str
            early_stopping_rounds=None,  # type: Optional[int]
            evals_result=None,  # type: Optional[Dict[Any, Any]]
            verbose_eval=True,  # type: Optional[bool]
            learning_rates=None,  # type: Optional[List[float]]
            keep_training_booster=False,  # type: Optional[bool]
            callbacks=None,  # type: Optional[List[Callable[..., Any]]]
            time_budget=None,  # type: Optional[int]
            sample_size=None,  # type: Optional[int]
            best_params=None,  # type: Optional[Dict[str, Any]]
            tuning_history=None,  # type: Optional[List[Dict[str, Any]]]
            verbosity=1,  # type: Optional[int]
    ):
        params = copy.deepcopy(params)
        args = [params, train_set]
        kwargs = dict(num_boost_round=num_boost_round,
                      valid_sets=valid_sets,
                      valid_names=valid_names,
                      fobj=fobj,
                      feval=feval,
                      feature_name=feature_name,
                      categorical_feature=categorical_feature,
                      early_stopping_rounds=early_stopping_rounds,
                      evals_result=evals_result,
                      verbose_eval=verbose_eval,
                      learning_rates=learning_rates,
                      keep_training_booster=keep_training_booster,
                      callbacks=callbacks,
                      time_budget=time_budget,
                      verbosity=verbosity,
                      sample_size=sample_size)  # type: Dict[str, Any]
        self._parse_args(*args, **kwargs)
        self.best_booster = None

        self.best_score = -np.inf if self.higher_is_better() else np.inf
        self.best_params = {} if best_params is None else best_params
        self.tuning_history = [] if tuning_history is None else tuning_history

        if early_stopping_rounds is None:
            self._suggest_early_stopping_rounds()
        if valid_sets is None:
            raise ValueError("`valid_sets` is required.")

    def _get_params(self):
        # type: () -> Dict[str, Any]

        params = copy.deepcopy(self.lgbm_params)
        params.update(self.best_params)
        return params

    def _parse_args(self, *args, **kwargs):
        # type: (Any, Any) -> None

        self.auto_options = {
            option_name: kwargs.get(option_name)
            for option_name in [
                'time_budget',
                'sample_size',
                'best_params',
                'tuning_history',
                'verbosity',
            ]
        }

        # Split options.
        for option_name in self.auto_options.keys():
            if option_name in kwargs:
                del kwargs[option_name]

        self.lgbm_params = args[0]
        self.train_set = args[1]
        self.train_subset = None  # Use for sampling.
        self.lgbm_kwargs = kwargs

        # Keep original kwargs.
        self.original_lgbm_kwargs = kwargs.copy()
        self.original_lgbm_params = self.lgbm_params.copy()

    def _suggest_early_stopping_rounds(self):
        # type: () -> int

        num_boost_round = self.lgbm_kwargs.get('num_boost_round', 1000)
        early_stopping_rounds = min(int(num_boost_round * 0.05), 50)
        return early_stopping_rounds

    def run(self):
        # type: () -> lgb.Booster
        """Perform the hyperparameter-tuning with given parameters.

        Returns:

            booster : Booster
                The trained Booster model.
        """
        # Surpress log messages.
        if self.auto_options['verbosity'] == 0:
            optuna.logging.disable_default_handler()
            self.lgbm_params['verbose'] = -1
            self.lgbm_params['seed'] = 111
            self.lgbm_kwargs['verbose_eval'] = False

        # Handling aliases.
        _handling_alias_parameters(self.lgbm_params)

        # Sampling.
        self.sample_train_set()

        # Tuning.
        time_budget = self.auto_options['time_budget']

        self.start_time = time.time()
        with _timer() as t:
            self.tune_feature_fraction()
            if time_budget is not None and time_budget > t.elapsed_secs():
                self.best_params.update(self._get_params())
                return self.best_booster

            self.tune_num_leaves()
            if time_budget is not None and time_budget > t.elapsed_secs():
                self.best_params.update(self._get_params())
                return self.best_booster

            self.tune_bagging()
            if time_budget is not None and time_budget > t.elapsed_secs():
                self.best_params.update(self._get_params())
                return self.best_booster

            self.tune_feature_fraction_stage2()
            if time_budget is not None and time_budget > t.elapsed_secs():
                self.best_params.update(self._get_params())
                return self.best_booster

            self.tune_regularization_factors()
            if time_budget is not None and time_budget > t.elapsed_secs():
                self.best_params.update(self._get_params())
                return self.best_booster

            self.tune_min_data_in_leaf()
            if time_budget is not None and time_budget > t.elapsed_secs():
                self.best_params.update(self._get_params())
                return self.best_booster

        self.best_params.update(self._get_params())
        return self.best_booster

    def sample_train_set(self):
        # type: () -> None
        """Make subset of `self.train_set` Dataset object."""

        if self.auto_options['sample_size'] is None:
            return

        self.train_set.construct()
        n_train_instance = self.train_set.get_label().shape[0]
        if n_train_instance > self.auto_options['sample_size']:
            offset = n_train_instance - self.auto_options['sample_size']
            idx_list = offset + np.arange(self.auto_options['sample_size'])
            self.train_subset = self.train_set.subset(idx_list)

    def tune_feature_fraction(self, n_trials=7):
        # type: (int) -> None

        param_name = 'feature_fraction'
        param_values = list(np.linspace(0.4, 1.0, n_trials))
        sampler = _GridSamplerUniform1D(param_name, param_values)
        self.tune_params([param_name], len(param_values), sampler)

    def tune_num_leaves(self, n_trials=20):
        # type: (int) -> None

        self.tune_params(['num_leaves'], n_trials, optuna.samplers.TPESampler())

    def tune_bagging(self, n_trials=10):
        # type: (int) -> None

        self.tune_params(['bagging_fraction', 'bagging_freq'],
                         n_trials,
                         optuna.samplers.TPESampler())

    def tune_feature_fraction_stage2(self, n_trials=6):
        # type: (int) -> None

        param_name = 'feature_fraction'
        param_values = list(np.linspace(
            self.lgbm_params[param_name] - 0.08,
            self.lgbm_params[param_name] + 0.08,
            n_trials))
        param_values = [val for val in param_values if val >= 0.4 and val <= 1.0]
        sampler = _GridSamplerUniform1D(param_name, param_values)
        self.tune_params([param_name], len(param_values), sampler)

    def tune_regularization_factors(self, n_trials=20):
        # type: (int) -> None

        self.tune_params(['lambda_l1', 'lambda_l2'], n_trials, optuna.samplers.TPESampler())

    def tune_min_data_in_leaf(self):
        # type: () -> None

        param_name = 'min_child_samples'
        param_values = [5, 10, 25, 50, 100]
        sampler = _GridSamplerUniform1D(param_name, param_values)
        self.tune_params([param_name], len(param_values), sampler)

    def tune_params(self, target_param_names, n_trials, sampler):
        # type: (List[str], int, optuna.samplers.BaseSampler) -> None

        pbar = tqdm.tqdm(total=n_trials, ascii=True)

        # Set current best parameters.
        self.lgbm_params.update(self.best_params)

        train_set = self.train_set
        if self.train_subset is not None:
            train_set = self.train_subset

        objective = OptunaObjective(
            target_param_names,
            self.lgbm_params,
            train_set,
            self.lgbm_kwargs,
            self.best_score,
            pbar=pbar,
        )
        study = optuna.create_study(
            direction='maximize' if self.higher_is_better() else 'minimize',
            sampler=sampler)
        study.optimize(objective, n_trials=n_trials, catch=())

        pbar.close()
        del pbar

        # Add tuning history.
        self.tuning_history += objective.report

        updated_params = {p: study.best_trial.params[p] for p in target_param_names}
        self.lgbm_params.update(updated_params)
        self.best_params.update(updated_params)

        if self.compare_validation_metrics(study.best_value, self.best_score):
            self.best_score = study.best_value
            self.best_booster = objective.best_booster



import abc
import six

from optuna.type_checking import TYPE_CHECKING

if TYPE_CHECKING:
    from optuna.structs import FrozenTrial  # NOQA
    from optuna.study import Study  # NOQA


@six.add_metaclass(abc.ABCMeta)
class BasePruner(object):
    """Base class for pruners."""

    @abc.abstractmethod
    def prune(self, study, trial):
        # type: (Study, FrozenTrial) -> bool
        """Judge whether the trial should be pruned at the given step.

        Note that this method is not supposed to be called by library users. Instead,
        :func:`optuna.trial.Trial.report` and :func:`optuna.trial.Trial.should_prune` provide
        user interfaces to implement pruning mechanism in an objective function.

        Args:
            study:
                Study object of the target study.
            trial:
                FrozenTrial object of the target trial.

        Returns:
            A boolean value representing whether the trial should be pruned.
        """

        raise NotImplementedError



from optuna.pruners.percentile import PercentilePruner


class MedianPruner(PercentilePruner):
    """Pruner using the median stopping rule.

    Prune if the trial's best intermediate result is worse than median of intermediate results of
    previous trials at the same step.

    Example:

        We minimize an objective function with the median stopping rule.

        .. code::

            >>> from optuna import create_study
            >>> from optuna.pruners import MedianPruner
            >>>
            >>> def objective(trial):
            >>>     ...
            >>>
            >>> study = create_study(pruner=MedianPruner())
            >>> study.optimize(objective)

    Args:
        n_startup_trials:
            Pruning is disabled until the given number of trials finish in the same study.
        n_warmup_steps:
            Pruning is disabled until the trial reaches the given number of step.
    """

    def __init__(self, n_startup_trials=5, n_warmup_steps=0):
        # type: (int, int) -> None

        super(MedianPruner, self).__init__(50.0, n_startup_trials, n_warmup_steps)



import pytest

import optuna
from optuna.storages import InMemoryStorage


@pytest.mark.parametrize('study_direction', ['minimize', 'maximize'])
def test_update_cache(study_direction):
    # type: (str) -> None

    storage = InMemoryStorage()

    assert storage.best_trial_id is None

    # If the direction is 'minimize', the objective function is weakly decreasing.
    # Otherwise, it is weakly increasing.
    sign = -1 if study_direction == 'minimize' else 1

    study = optuna.create_study(storage=storage, direction=study_direction)
    study.optimize(lambda trial: sign * trial.number, n_trials=1)
    assert storage.best_trial_id == 0

    study.optimize(lambda trial: sign * trial.number, n_trials=1)
    assert storage.best_trial_id == 1

    # The objective value is equal to the best value.
    study.optimize(lambda trial: sign * (trial.number - 1), n_trials=1)
    assert storage.best_trial_id == 1

    # The objective value is inferior to the best value.
    study.optimize(lambda trial: sign * (trial.number - 2), n_trials=1)
    assert storage.best_trial_id == 1


def test_update_cache_none_value():
    # type: () -> None

    storage = InMemoryStorage()

    study = optuna.create_study(storage=storage)
    study.optimize(lambda trial: -1 * trial.number, n_trials=1)
    assert storage.best_trial_id == 0

    # The objective value is None.
    study.optimize(lambda trial: None, n_trials=1)  # type: ignore
    assert storage.best_trial_id == 0



try:
    import bokeh.command.bootstrap
    import bokeh.document  # NOQA
    import bokeh.layouts
    import bokeh.models
    import bokeh.models.widgets
    import bokeh.plotting
    import bokeh.themes
    import tornado.gen
    _available = True
except ImportError as e:
    _available = False
    _import_error = e

import collections
import numpy as np
import threading
import time

import optuna.logging
import optuna.structs
import optuna.study
from optuna import type_checking

if type_checking.TYPE_CHECKING:
    from typing import Any  # NOQA
    from typing import Dict  # NOQA
    from typing import List  # NOQA
    from typing import Optional  # NOQA

_mode = None  # type: Optional[str]
_study = None  # type: Optional[optuna.study.Study]

_HEADER_FORMAT = '''
<style>
body {{
    margin: 20px;
}}
h1, p {{
    margin: 10px 0px;
}}
</style>

<h1>Optuna Dashboard (Beta)</h1>
<p>
<b>Study name:</b> {study_name}<br>
</p>
'''

_DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'

if _available:

    class _CompleteTrialsWidget(object):
        def __init__(self, trials, direction):
            # type: (List[optuna.structs.FrozenTrial], optuna.structs.StudyDirection) -> None

            complete_trials = [
                trial for trial in trials if trial.state == optuna.structs.TrialState.COMPLETE
            ]
            self.trial_ids = set([trial.trial_id for trial in complete_trials])

            self.direction = direction
            values = [trial.value for trial in complete_trials]
            if direction == optuna.structs.StudyDirection.MINIMIZE:
                best_values = np.minimum.accumulate(values, axis=0)
            else:
                best_values = np.maximum.accumulate(values, axis=0)

            self.cds = bokeh.models.ColumnDataSource({
                '#': list(range(len(complete_trials))),
                'value': values,
                'best_value': best_values,
            })

            self.best_value = best_values[-1] if complete_trials else np.inf

        def create_figure(self):
            # type: () -> bokeh.plotting.Figure

            figure = bokeh.plotting.figure(height=150)
            figure.circle(x='#', y='value', source=self.cds, alpha=0.3, color='navy')
            figure.line(x='#', y='best_value', source=self.cds, color='firebrick')
            figure.xaxis[0].axis_label = 'Number of Trials'
            figure.yaxis[0].axis_label = 'Objective Value'
            return figure

        def update(self, new_trials):
            # type: (List[optuna.structs.FrozenTrial]) -> None

            stream_dict = collections.defaultdict(list)  # type: Dict[str, List[Any]]

            for trial in new_trials:
                if trial.state != optuna.structs.TrialState.COMPLETE:
                    continue
                if trial.trial_id in self.trial_ids:
                    continue
                stream_dict['#'].append(len(self.trial_ids))
                stream_dict['value'].append(trial.value)
                if self.direction == optuna.structs.StudyDirection.MINIMIZE:
                    self.best_value = min(self.best_value, trial.value)
                else:
                    self.best_value = max(self.best_value, trial.value)
                stream_dict['best_value'].append(self.best_value)
                self.trial_ids.add(trial.trial_id)

            if stream_dict:
                self.cds.stream(stream_dict)

    class _AllTrialsWidget(object):
        def __init__(self, trials):
            # type: (List[optuna.structs.FrozenTrial]) -> None

            self.cds = bokeh.models.ColumnDataSource(self.trials_to_dict(trials))

        def create_table(self):
            # type: () -> bokeh.models.widgets.DataTable

            return bokeh.models.widgets.DataTable(
                source=self.cds,
                columns=[
                    bokeh.models.widgets.TableColumn(field=field, title=field) for field in
                    ['number', 'state', 'value', 'params', 'datetime_start', 'datetime_complete']
                ])

        def update(
                self,
                old_trials,  # type: List[optuna.structs.FrozenTrial]
                new_trials,  # type: List[optuna.structs.FrozenTrial]
        ):
            # type: (...) -> None

            modified_indices = []
            modified_trials = []
            for i, old_trial in enumerate(old_trials):
                new_trial = new_trials[i]
                if old_trial != new_trial:
                    modified_indices.append(i)
                    modified_trials.append(new_trial)

            patch_dict = self.trials_to_dict(modified_trials)
            patch_dict = {k: list(zip(modified_indices, v)) for k, v in patch_dict.items()}
            self.cds.patch(patch_dict)

            self.cds.stream(self.trials_to_dict(new_trials[len(old_trials):]))

        @staticmethod
        def trials_to_dict(trials):
            # type: (List[optuna.structs.FrozenTrial]) -> Dict[str, List[Any]]

            return {
                'number': [trial.number for trial in trials],
                'state': [trial.state.name for trial in trials],
                'value': [trial.value for trial in trials],
                'params': [str(trial.params) for trial in trials],
                'datetime_start': [
                    trial.datetime_start.strftime(_DATETIME_FORMAT)
                    if trial.datetime_start is not None else None for trial in trials
                ],
                'datetime_complete': [
                    trial.datetime_complete.strftime(_DATETIME_FORMAT)
                    if trial.datetime_complete is not None else None for trial in trials
                ],
            }

    class _DashboardApp(object):
        def __init__(self, study, launch_update_thread):
            # type: (optuna.study.Study, bool) -> None

            self.study = study
            self.launch_update_thread = launch_update_thread
            self.lock = threading.Lock()

        def __call__(self, doc):
            # type: (bokeh.document.Document) -> None

            self.doc = doc
            self.current_trials = \
                self.study.trials  # type: Optional[List[optuna.structs.FrozenTrial]]
            self.new_trials = None  # type: Optional[List[optuna.structs.FrozenTrial]]
            self.complete_trials_widget = _CompleteTrialsWidget(
                self.current_trials, self.study.direction)
            self.all_trials_widget = _AllTrialsWidget(self.current_trials)

            self.doc.title = 'Optuna Dashboard (Beta)'
            header = _HEADER_FORMAT.format(study_name=self.study.study_name)
            self.doc.add_root(
                bokeh.layouts.layout([[bokeh.models.widgets.Div(text=header)],
                                      [self.complete_trials_widget.create_figure()],
                                      [self.all_trials_widget.create_table()]],
                                     sizing_mode='scale_width'))

            if self.launch_update_thread:
                thread = threading.Thread(target=self.thread_loop)
                thread.daemon = True
                thread.start()

        def thread_loop(self):
            # type: () -> None

            while True:
                time.sleep(1)
                new_trials = self.study.trials
                with self.lock:
                    need_to_add_callback = (self.new_trials is None)
                    self.new_trials = new_trials
                    if need_to_add_callback:
                        self.doc.add_next_tick_callback(self.update_callback)

        @tornado.gen.coroutine
        def update_callback(self):
            # type: () -> None

            with self.lock:
                current_trials = self.current_trials
                new_trials = self.new_trials
                self.current_trials = self.new_trials
                self.new_trials = None

            assert current_trials is not None
            assert new_trials is not None
            self.complete_trials_widget.update(new_trials)
            self.all_trials_widget.update(current_trials, new_trials)


def _check_bokeh_availability():
    # type: () -> None

    if not _available:
        raise ImportError(
            'Bokeh is not available. Please install Bokeh to use the dashboard. '
            'Bokeh can be installed by executing `$ pip install bokeh`. '
            'For further information, please refer to the installation guide of Bokeh. '
            '(The actual import error is as follows: ' + str(_import_error) + ')')


def _show_experimental_warning():
    # type: () -> None

    logger = optuna.logging.get_logger(__name__)
    logger.warning('Optuna dashboard is still highly experimental. Please use with caution!')


def _get_this_source_path():
    # type: () -> str

    path = __file__

    # Sometimes __file__ points to a *.pyc file, but Bokeh doesn't accept it.
    if path.endswith('.pyc'):
        path = path[:-1]
    return path


def serve(study, bokeh_allow_websocket_origins=None):
    # type: (optuna.study.Study, Optional[List[str]]) -> None

    global _mode, _study

    _check_bokeh_availability()
    _show_experimental_warning()

    # We want to pass the mode (launching a server? or, just writing an HTML?) and a target study
    # to our Bokeh app. Unfortunately, as we are using `bokeh.command.bootstrap.main` to launch
    # our Bokeh app, we cannot directly pass Python objects to it. Therefore, we have no choice but
    # to use global variables to pass them.
    _mode = 'serve'
    _study = study

    # TODO(akiba): Stop using Bokeh's CLI entry point, and start the HTTP server by ourselves.

    # This is not a very clean way to launch Bokeh server.
    # Another seemingly better way is to
    # instantiate and launch `bokeh.server.server.Server` by ourselves. However, in this way,
    # for some reason, we found that the CDS update is not reflected to browsers, at least on Bokeh
    # version 0.12.15. In addition, we will need to do many configuration to servers, which can be
    # done automatically with the following one line. So, for now, we decided to use this way.
    command = ['bokeh', 'serve', '--show', _get_this_source_path()]
    if bokeh_allow_websocket_origins is not None:
        for bokeh_allow_websocket_origin in bokeh_allow_websocket_origins:
            command.extend(['--allow-websocket-origin', bokeh_allow_websocket_origin])
    bokeh.command.bootstrap.main(command)


def write(study, out_path):
    # type: (optuna.study.Study, str) -> None

    global _mode, _study

    _check_bokeh_availability()
    _show_experimental_warning()

    _mode = 'html'
    _study = study
    bokeh.command.bootstrap.main(['bokeh', 'html', _get_this_source_path(), '-o', out_path])


def _run():
    # type: () -> None

    # Please note that `_study` and `optuna.dashboard._study` are different here. Here, this module
    # is loaded inside Bokeh, and thus it is not `optuna.dashboard`, but `bk_script_????`.
    study = optuna.dashboard._study
    mode = optuna.dashboard._mode

    assert study is not None
    app = _DashboardApp(study, launch_update_thread=(mode == 'serve'))
    doc = bokeh.plotting.curdoc()
    app(doc)


if __name__.startswith('bk_script_'):
    # Here, this module is loaded inside Bokeh. Therefore, we should launch the Bokeh app.
    _run()



from __future__ import absolute_import

import optuna
from optuna import type_checking

try:
    import chainer
    from chainer.training.extension import Extension
    from chainer.training import triggers
    _available = True
except ImportError as e:
    _import_error = e
    # ChainerPruningExtension is disabled because Chainer is not available.
    _available = False
    # This alias is required to avoid ImportError at ChainerPruningExtension definition.
    Extension = object

if type_checking.TYPE_CHECKING:
    from typing import Tuple
    from typing import Union

    TriggerType = Union[Tuple[(int,
                               str)], triggers.IntervalTrigger, triggers.ManualScheduleTrigger]


class ChainerPruningExtension(Extension):
    """Chainer extension to prune unpromising trials.

    Example:

        Add a pruning extension which observes validation losses to
        `Chainer Trainer <https://docs.chainer.org/en/stable/reference/generated/
        chainer.training.Trainer.html>`_.

        .. code::

            trainer.extend(
                ChainerPruningExtension(trial, 'validation/main/loss', (1, 'epoch')))

    Args:
        trial:
            A :class:`~optuna.trial.Trial` corresponding to the current evaluation of the
            objective function.
        observation_key:
            An evaluation metric for pruning, e.g., ``main/loss`` and
            ``validation/main/accuracy``. Please refer to
            `chainer.Reporter reference <https://docs.chainer.org/en/stable/reference/
            util/generated/chainer.Reporter.html>`_ for further details.
        pruner_trigger:
            A trigger to execute pruning. ``pruner_trigger`` is an instance of
            `IntervalTrigger <https://docs.chainer.org/en/stable/reference/generated/
            chainer.training.triggers.IntervalTrigger.html>`_ or
            `ManualScheduleTrigger <https://docs.chainer.org/en/stable/reference/generated/
            chainer.training.triggers.ManualScheduleTrigger.html>`_. `IntervalTrigger <https://
            docs.chainer.org/en/stable/reference/generated/chainer.training.triggers.
            IntervalTrigger.html>`_ can be specified by a tuple of the interval length and its
            unit like ``(1, 'epoch')``.
    """

    def __init__(self, trial, observation_key, pruner_trigger):
        # type: (optuna.trial.Trial, str, TriggerType) -> None

        _check_chainer_availability()

        self.trial = trial
        self.observation_key = observation_key
        self.pruner_trigger = chainer.training.get_trigger(pruner_trigger)
        if not (isinstance(self.pruner_trigger, triggers.IntervalTrigger)
                or isinstance(self.pruner_trigger, triggers.ManualScheduleTrigger)):
            pruner_type = type(self.pruner_trigger)
            raise TypeError("Invalid trigger class: " + str(pruner_type) + "\n"
                            "Pruner trigger is supposed to be an instance of "
                            "IntervalTrigger or ManualScheduleTrigger.")

    @staticmethod
    def _get_float_value(observation_value):
        # type: (Union[float, chainer.Variable]) -> float

        _check_chainer_availability()

        if isinstance(observation_value, chainer.Variable):
            observation_value = observation_value.data

        try:
            observation_value = float(observation_value)
        except TypeError:
            raise TypeError(
                'Type of observation value is not supported by ChainerPruningExtension.\n'
                '{} cannot be casted to float.'.format(type(observation_value)))

        return observation_value

    def _observation_exists(self, trainer):
        # type: (chainer.training.Trainer) -> bool

        return self.pruner_trigger(trainer) and self.observation_key in trainer.observation

    def __call__(self, trainer):
        # type: (chainer.training.Trainer) -> None

        if not self._observation_exists(trainer):
            return

        current_score = self._get_float_value(trainer.observation[self.observation_key])
        current_step = getattr(trainer.updater, self.pruner_trigger.unit)
        self.trial.report(current_score, step=current_step)
        if self.trial.should_prune():
            message = "Trial was pruned at {} {}.".format(self.pruner_trigger.unit, current_step)
            raise optuna.structs.TrialPruned(message)


def _check_chainer_availability():
    # type: () -> None

    if not _available:
        raise ImportError(
            'Chainer is not available. Please install Chainer to use this feature. '
            'Chainer can be installed by executing `$ pip install chainer`. '
            'For further information, please refer to the installation guide of Chainer. '
            '(The actual import error is as follows: ' + str(_import_error) + ')')



from optuna.pruners import BasePruner
from optuna import type_checking

if type_checking.TYPE_CHECKING:
    from optuna import structs  # NOQA
    from optuna.study import Study  # NOQA


class NopPruner(BasePruner):
    """Pruner which never prunes trials.

    Example:

        .. code::

            >>> from optuna import create_study
            >>> from optuna.pruners import NopPruner
            >>>
            >>> def objective(trial):
            >>>     ...
            >>>
            >>> study = create_study(pruner=NopPruner())
            >>> study.optimize(objective)
    """

    def prune(self, study, trial):
        # type: (Study, structs.FrozenTrial) -> bool
        """Please consult the documentation for :func:`BasePruner.prune`."""

        return False



from __future__ import absolute_import

from logging import DEBUG
from logging import INFO
from logging import WARNING
from numbers import Number
from time import time

import numpy as np

try:
    from sklearn.base import BaseEstimator
    from sklearn.base import clone
    from sklearn.base import is_classifier
    from sklearn.metrics.scorer import check_scoring
    from sklearn.model_selection._validation import _index_param_value
    from sklearn.model_selection import BaseCrossValidator  # NOQA
    from sklearn.model_selection import check_cv
    from sklearn.model_selection import cross_validate
    from sklearn.utils import check_random_state
    from sklearn.utils.metaestimators import _safe_split
    from sklearn.utils import safe_indexing as sklearn_safe_indexing
    from sklearn.utils.validation import _num_samples
    from sklearn.utils.validation import check_is_fitted

    _available = True

except ImportError as e:
    BaseEstimator = object

    _import_error = e
    _available = False

from optuna import distributions
from optuna import logging
from optuna import samplers  # NOQA
from optuna import structs
from optuna import study as study_module
from optuna import trial as trial_module  # NOQA
from optuna import type_checking

if type_checking.TYPE_CHECKING:
    import pandas as pd  # NOQA
    from scipy.sparse import spmatrix  # NOQA
    from typing import Any  # NOQA
    from typing import Callable  # NOQA
    from typing import Dict  # NOQA
    from typing import List  # NOQA
    from typing import Mapping  # NOQA
    from typing import Optional  # NOQA
    from typing import Union  # NOQA

    OneDimArrayLikeType = Union[List[float], np.ndarray, pd.Series]
    TwoDimArrayLikeType = \
        Union[List[List[float]], np.ndarray, pd.DataFrame, spmatrix]

logger = logging.get_logger(__name__)


def _check_sklearn_availability():
    # type: () -> None

    if not _available:
        raise ImportError(
            'scikit-learn is not available. Please install scikit-learn to '
            'use this feature. scikit-learn can be installed by executing '
            '`$ pip install scikit-learn>=0.19.0`. For further information, '
            'please refer to the installation guide of scikit-learn. (The '
            'actual import error is as follows: ' + str(_import_error) + ')'
        )


def safe_indexing(
    X,  # type: Union[OneDimArrayLikeType, TwoDimArrayLikeType]
    indices  # type: OneDimArrayLikeType
):
    # type: (...) -> Union[OneDimArrayLikeType, TwoDimArrayLikeType]
    if X is None:
        return X

    return sklearn_safe_indexing(X, indices)


class _Objective(object):
    """Callable that implements objective function.

    Args:
        estimator:
            Object to use to fit the data. This is assumed to implement the
            scikit-learn estimator interface. Either this needs to provide
            ``score``, or ``scoring`` must be passed.

        param_distributions:
            Dictionary where keys are parameters and values are distributions.
            Distributions are assumed to implement the optuna distribution
            interface.

        X:
            Training data.

        y:
            Target variable.

        cv:
            Cross-validation strategy.

        enable_pruning:
            If :obj:`True`, pruning is performed in the case where the
            underlying estimator supports ``partial_fit``.

        error_score:
            Value to assign to the score if an error occurs in fitting. If
            'raise', the error is raised. If numeric,
            ``sklearn.exceptions.FitFailedWarning`` is raised. This does not
            affect the refit step, which will always raise the error.

        fit_params:
            Parameters passed to ``fit`` one the estimator.

        groups:
            Group labels for the samples used while splitting the dataset into
            train/test set.

        max_iter:
            Maximum number of epochs. This is only used if the underlying
            estimator supports ``partial_fit``.

        return_train_score:
            If :obj:`True`, training scores will be included. Computing
            training scores is used to get insights on how different
            hyperparameter settings impact the overfitting/underfitting
            trade-off. However computing training scores can be
            computationally expensive and is not strictly required to select
            the hyperparameters that yield the best generalization
            performance.

        scoring:
            Scorer function.
    """

    def __init__(
        self,
        estimator,  # type: BaseEstimator
        param_distributions,  # type: Mapping[str, distributions.BaseDistribution]
        X,  # type: TwoDimArrayLikeType
        y,  # type: Optional[Union[OneDimArrayLikeType, TwoDimArrayLikeType]]
        cv,  # type: BaseCrossValidator
        enable_pruning,  # type: bool
        error_score,  # type: Union[float, str]
        fit_params,  # type: Dict[str, Any]
        groups,  # type: Optional[OneDimArrayLikeType]
        max_iter,  # type: int
        return_train_score,  # type: bool
        scoring  # type: Callable[..., float]
    ):
        # type: (...) -> None

        self.cv = cv
        self.enable_pruning = enable_pruning
        self.error_score = error_score
        self.estimator = estimator
        self.fit_params = fit_params
        self.groups = groups
        self.max_iter = max_iter
        self.param_distributions = param_distributions
        self.return_train_score = return_train_score
        self.scoring = scoring
        self.X = X
        self.y = y

    def __call__(self, trial):
        # type: (trial_module.Trial) -> float

        estimator = clone(self.estimator)
        params = self._get_params(trial)

        estimator.set_params(**params)

        if self.enable_pruning:
            scores = self._cross_validate_with_pruning(trial, estimator)
        else:
            scores = cross_validate(
                estimator,
                self.X,
                self.y,
                cv=self.cv,
                error_score=self.error_score,
                fit_params=self.fit_params,
                groups=self.groups,
                return_train_score=self.return_train_score,
                scoring=self.scoring
            )

        self._store_scores(trial, scores)

        return trial.user_attrs['mean_test_score']

    def _cross_validate_with_pruning(
        self,
        trial,  # type: trial_module.Trial
        estimator  # type: BaseEstimator
    ):
        # type: (...) -> Dict[str, OneDimArrayLikeType]

        if is_classifier(estimator):
            partial_fit_params = self.fit_params.copy()
            classes = np.unique(self.y)

            partial_fit_params.setdefault('classes', classes)

        else:
            partial_fit_params = self.fit_params

        n_splits = self.cv.get_n_splits(self.X, self.y, groups=self.groups)
        estimators = [clone(estimator) for _ in range(n_splits)]
        scores = {
            'fit_time': np.zeros(n_splits),
            'score_time': np.zeros(n_splits),
            'test_score': np.empty(n_splits)
        }

        if self.return_train_score:
            scores['train_score'] = np.empty(n_splits)

        for step in range(self.max_iter):
            for i, (train, test) in enumerate(
                self.cv.split(self.X, self.y, groups=self.groups)
            ):
                out = self._partial_fit_and_score(
                    estimators[i],
                    train,
                    test,
                    partial_fit_params
                )

                if self.return_train_score:
                    scores['train_score'][i] = out.pop(0)

                scores['test_score'][i] = out[0]
                scores['fit_time'][i] += out[1]
                scores['score_time'][i] += out[2]

            intermediate_value = np.nanmean(scores['test_score'])

            trial.report(intermediate_value, step=step)

            if trial.should_prune():
                self._store_scores(trial, scores)

                raise structs.TrialPruned(
                    'trial was pruned at iteration {}.'.format(step)
                )

        return scores

    def _get_params(self, trial):
        # type: (trial_module.Trial) -> Dict[str, Any]

        return {
            name: trial._suggest(
                name, distribution
            ) for name, distribution in self.param_distributions.items()
        }

    def _partial_fit_and_score(
        self,
        estimator,  # type: BaseEstimator
        train,  # type: List[int]
        test,  # type: List[int]
        partial_fit_params  # type: Dict[str, Any]
    ):
        # type: (...) -> List[float]

        X_train, y_train = _safe_split(estimator, self.X, self.y, train)
        X_test, y_test = _safe_split(
            estimator,
            self.X,
            self.y,
            test,
            train_indices=train
        )

        start_time = time()

        try:
            estimator.partial_fit(X_train, y_train, **partial_fit_params)

        except Exception as e:
            if self.error_score == 'raise':
                raise e

            elif isinstance(self.error_score, Number):
                fit_time = time() - start_time
                test_score = self.error_score
                score_time = 0.0

                if self.return_train_score:
                    train_score = self.error_score

            else:
                raise ValueError(
                    'error_score must be \'raise\' or numeric.'
                )

        else:
            fit_time = time() - start_time
            test_score = self.scoring(estimator, X_test, y_test)
            score_time = time() - fit_time - start_time

            if self.return_train_score:
                train_score = self.scoring(estimator, X_train, y_train)

        ret = [test_score, fit_time, score_time]

        if self.return_train_score:
            ret.insert(0, train_score)

        return ret

    def _store_scores(self, trial, scores):
        # type: (trial_module.Trial, Dict[str, OneDimArrayLikeType]) -> None

        for name, array in scores.items():
            if name in ['test_score', 'train_score']:
                for i, score in enumerate(array):
                    trial.set_user_attr('split{}_{}'.format(i, name), score)

            trial.set_user_attr('mean_{}'.format(name), np.nanmean(array))
            trial.set_user_attr('std_{}'.format(name), np.nanstd(array))


class OptunaSearchCV(BaseEstimator):
    """Hyperparameter search with cross-validation.

    .. warning::

        This feature is experimental. The interface may be changed in the future.

    Args:
        estimator:
            Object to use to fit the data. This is assumed to implement the
            scikit-learn estimator interface. Either this needs to provide
            ``score``, or ``scoring`` must be passed.

        param_distributions:
            Dictionary where keys are parameters and values are distributions.
            Distributions are assumed to implement the optuna distribution
            interface.

        cv:
            Cross-validation strategy. Possible inputs for cv are:

            - integer to specify the number of folds in a CV splitter,
            - a CV splitter,
            - an iterable yielding (train, test) splits as arrays of indices.

            For integer, if :obj:`estimator` is a classifier and :obj:`y` is
            either binary or multiclass,
            ``sklearn.model_selection.StratifiedKFold`` is used. otherwise,
            ``sklearn.model_selection.KFold`` is used.

        enable_pruning:
            If :obj:`True`, pruning is performed in the case where the
            underlying estimator supports ``partial_fit``.

        error_score:
            Value to assign to the score if an error occurs in fitting. If
            'raise', the error is raised. If numeric,
            ``sklearn.exceptions.FitFailedWarning`` is raised. This does not
            affect the refit step, which will always raise the error.

        max_iter:
            Maximum number of epochs. This is only used if the underlying
            estimator supports ``partial_fit``.

        n_jobs:
            Number of parallel jobs. :obj:`-1` means using all processors.

        n_trials:
            Number of trials. If :obj:`None`, there is no limitation on the
            number of trials. If :obj:`timeout` is also set to :obj:`None`,
            the study continues to create trials until it receives a
            termination signal such as Ctrl+C or SIGTERM. This trades off
            runtime vs quality of the solution.

        random_state:
            Seed of the pseudo random number generator. If int, this is the
            seed used by the random number generator. If
            ``numpy.random.RandomState`` object, this is the random number
            generator. If :obj:`None`, the global random state from
            ``numpy.random`` is used.

        refit:
            If :obj:`True`, refit the estimator with the best found
            hyperparameters. The refitted estimator is made available at the
            ``best_estimator_`` attribute and permits using ``predict``
            directly.

        return_train_score:
            If :obj:`True`, training scores will be included. Computing
            training scores is used to get insights on how different
            hyperparameter settings impact the overfitting/underfitting
            trade-off. However computing training scores can be
            computationally expensive and is not strictly required to select
            the hyperparameters that yield the best generalization
            performance.

        scoring:
            String or callable to evaluate the predictions on the test data.
            If :obj:`None`, ``score`` on the estimator is used.

        study:
            Study corresponds to the optimization task. If :obj:`None`, a new
            study is created.

        subsample:
            Proportion of samples that are used during hyperparameter search.

            - If int, then draw ``subsample`` samples.
            - If float, then draw ``subsample`` * ``X.shape[0]`` samples.

        timeout:
            Time limit in seconds for the search of appropriate models. If
            :obj:`None`, the study is executed without time limitation. If
            :obj:`n_trials` is also set to :obj:`None`, the study continues to
            create trials until it receives a termination signal such as
            Ctrl+C or SIGTERM. This trades off runtime vs quality of the
            solution.

        verbose:
            Verbosity level. The higher, the more messages.

    Attributes:
        best_estimator_:
            Estimator that was chosen by the search. This is present only if
            ``refit`` is set to :obj:`True`.

        n_splits_:
            Number of cross-validation splits.

        refit_time_:
            Time for refitting the best estimator. This is present only if
            ``refit`` is set to :obj:`True`.

        sample_indices_:
            Indices of samples that are used during hyperparameter search.

        scorer_:
            Scorer function.

        study_:
            Actual study.

    Examples:
        >>> import optuna
        >>> from sklearn.datasets import load_iris
        >>> from sklearn.svm import SVC
        >>> clf = SVC(gamma='auto')
        >>> param_distributions = {
        ...     'C': optuna.distributions.LogUniformDistribution(1e-10, 1e+10)
        ... }
        >>> optuna_search = optuna.integration.OptunaSearchCV(
        ...     clf,
        ...     param_distributions
        ... )
        >>> X, y = load_iris(return_X_y=True)
        >>> optuna_search.fit(X, y) # doctest: +ELLIPSIS
        OptunaSearchCV(...)
        >>> y_pred = optuna_search.predict(X)
    """

    _required_parameters = ['estimator', 'param_distributions']

    @property
    def _estimator_type(self):
        # type: () -> str

        return self.estimator._estimator_type

    @property
    def best_index_(self):
        # type: () -> int
        """Index which corresponds to the best candidate parameter setting."""

        df = self.trials_dataframe()

        return df['value'].idxmin()

    @property
    def best_params_(self):
        # type: () -> Dict[str, Any]
        """Parameters of the best trial in the :class:`~optuna.study.Study`."""

        self._check_is_fitted()

        return self.study_.best_params

    @property
    def best_score_(self):
        # type: () -> float
        """Mean cross-validated score of the best estimator."""

        self._check_is_fitted()

        return self.study_.best_value

    @property
    def best_trial_(self):
        # type: () -> structs.FrozenTrial
        """Best trial in the :class:`~optuna.study.Study`."""

        self._check_is_fitted()

        return self.study_.best_trial

    @property
    def classes_(self):
        # type: () -> OneDimArrayLikeType
        """Class labels."""

        self._check_is_fitted()

        return self.best_estimator_.classes_

    @property
    def n_trials_(self):
        # type: () -> int
        """Actual number of trials."""

        return len(self.trials_)

    @property
    def trials_(self):
        # type: () -> List[structs.FrozenTrial]
        """All trials in the :class:`~optuna.study.Study`."""

        self._check_is_fitted()

        return self.study_.trials

    @property
    def user_attrs_(self):
        # type: () -> Dict[str, Any]
        """User attributes in the :class:`~optuna.study.Study`."""

        self._check_is_fitted()

        return self.study_.user_attrs

    @property
    def decision_function(self):
        # type: () -> Callable[..., Union[OneDimArrayLikeType, TwoDimArrayLikeType]]
        """Call ``decision_function`` on the best estimator.

        This is available only if the underlying estimator supports
        ``decision_function`` and ``refit`` is set to :obj:`True`.
        """

        self._check_is_fitted()

        return self.best_estimator_.decision_function

    @property
    def inverse_transform(self):
        # type: () -> Callable[..., TwoDimArrayLikeType]
        """Call ``inverse_transform`` on the best estimator.

        This is available only if the underlying estimator supports
        ``inverse_transform`` and ``refit`` is set to :obj:`True`.
        """

        self._check_is_fitted()

        return self.best_estimator_.inverse_transform

    @property
    def predict(self):
        # type: () -> Callable[..., Union[OneDimArrayLikeType, TwoDimArrayLikeType]]
        """Call ``predict`` on the best estimator.

        This is available only if the underlying estimator supports ``predict``
        and ``refit`` is set to :obj:`True`.
        """

        self._check_is_fitted()

        return self.best_estimator_.predict

    @property
    def predict_log_proba(self):
        # type: () -> Callable[..., TwoDimArrayLikeType]
        """Call ``predict_log_proba`` on the best estimator.

        This is available only if the underlying estimator supports
        ``predict_log_proba`` and ``refit`` is set to :obj:`True`.
        """

        self._check_is_fitted()

        return self.best_estimator_.predict_log_proba

    @property
    def predict_proba(self):
        # type: () -> Callable[..., TwoDimArrayLikeType]
        """Call ``predict_proba`` on the best estimator.

        This is available only if the underlying estimator supports
        ``predict_proba`` and ``refit`` is set to :obj:`True`.
        """

        self._check_is_fitted()

        return self.best_estimator_.predict_proba

    @property
    def score_samples(self):
        # type: () -> Callable[..., OneDimArrayLikeType]
        """Call ``score_samples`` on the best estimator.

        This is available only if the underlying estimator supports
        ``score_samples`` and ``refit`` is set to :obj:`True`.
        """

        self._check_is_fitted()

        return self.best_estimator_.score_samples

    @property
    def set_user_attr(self):
        # type: () -> Callable[..., None]
        """Call ``set_user_attr`` on the :class:`~optuna.study.Study`."""

        self._check_is_fitted()

        return self.study_.set_user_attr

    @property
    def transform(self):
        # type: () -> Callable[..., TwoDimArrayLikeType]
        """Call ``transform`` on the best estimator.

        This is available only if the underlying estimator supports
        ``transform`` and ``refit`` is set to :obj:`True`.
        """

        self._check_is_fitted()

        return self.best_estimator_.transform

    @property
    def trials_dataframe(self):
        # type: () -> Callable[..., pd.DataFrame]
        """Call ``trials_dataframe`` on the :class:`~optuna.study.Study`."""

        self._check_is_fitted()

        return self.study_.trials_dataframe

    def __init__(
        self,
        estimator,  # type: BaseEstimator
        param_distributions,  # type: Mapping[str, distributions.BaseDistribution]
        cv=5,  # type: Optional[Union[BaseCrossValidator, int]]
        enable_pruning=False,  # type: bool
        error_score=np.nan,  # type: Union[float, str]
        max_iter=1000,  # type: int
        n_jobs=1,  # type: int
        n_trials=10,  # type: int
        random_state=None,  # type: Optional[Union[int, np.random.RandomState]]
        refit=True,  # type: bool
        return_train_score=False,  # type: bool
        scoring=None,  # type: Optional[Union[Callable[..., float], str]]
        study=None,  # type: Optional[study_module.Study]
        subsample=1.0,  # type: Union[float, int]
        timeout=None,  # type: Optional[float]
        verbose=0  # type: int
    ):
        # type: (...) -> None

        _check_sklearn_availability()

        self.cv = cv
        self.enable_pruning = enable_pruning
        self.error_score = error_score
        self.estimator = estimator
        self.max_iter = max_iter
        self.n_trials = n_trials
        self.n_jobs = n_jobs
        self.param_distributions = param_distributions
        self.random_state = random_state
        self.refit = refit
        self.return_train_score = return_train_score
        self.scoring = scoring
        self.study = study
        self.subsample = subsample
        self.timeout = timeout
        self.verbose = verbose

    def _check_is_fitted(self):
        # type: () -> None

        attributes = [
            'n_splits_',
            'sample_indices_',
            'scorer_',
            'study_'
        ]

        if self.refit:
            attributes += ['best_estimator_', 'refit_time_']

        check_is_fitted(self, attributes)

    def _check_params(self):
        # type: () -> None

        if not hasattr(self.estimator, 'fit'):
            raise ValueError(
                'estimator must be a scikit-learn estimator.'
            )

        if type(self.param_distributions) is not dict:
            raise ValueError('param_distributions must be a dictionary.')

        for name, distribution in self.param_distributions.items():
            if not isinstance(distribution, distributions.BaseDistribution):
                raise ValueError(
                    'Value of {} must be a optuna distribution.'.format(name)
                )

        if self.enable_pruning and not hasattr(self.estimator, 'partial_fit'):
            raise ValueError('estimator must support partial_fit.')

        if self.max_iter <= 0:
            raise ValueError(
                'max_iter must be > 0, got {}.'.format(self.max_iter)
            )

        if self.study is not None \
                and self.study.direction != structs.StudyDirection.MAXIMIZE:
            raise ValueError('direction of study must be \'maximize\'.')

    def _more_tags(self):
        # type: () -> Dict[str, bool]

        return {'non_deterministic': True, 'no_validation': True}

    def _refit(
        self,
        X,  # type: TwoDimArrayLikeType
        y=None,  # type: Optional[Union[OneDimArrayLikeType, TwoDimArrayLikeType]]
        **fit_params  # type: Any
    ):
        # type: (...) -> 'OptunaSearchCV'

        n_samples = _num_samples(X)

        self.best_estimator_ = clone(self.estimator)

        try:
            self.best_estimator_.set_params(**self.study_.best_params)
        except ValueError as e:
            logger.exception(e)

        logger.info(
            'Refitting the estimator using {} samples...'.format(n_samples)
        )

        start_time = time()

        self.best_estimator_.fit(X, y, **fit_params)

        self.refit_time_ = time() - start_time

        logger.info(
            'Finished refitting! '
            '(elapsed time: {:.3f} sec.)'.format(self.refit_time_)
        )

        return self

    def fit(
        self,
        X,  # type: TwoDimArrayLikeType
        y=None,  # type: Optional[Union[OneDimArrayLikeType, TwoDimArrayLikeType]]
        groups=None,  # type: Optional[OneDimArrayLikeType]
        **fit_params  # type: Any
    ):
        # type: (...) -> 'OptunaSearchCV'
        """Run fit with all sets of parameters.

        Args:
            X:
                Training data.

            y:
                Target variable.

            groups:
                Group labels for the samples used while splitting the dataset
                into train/test set.

            **fit_params:
                Parameters passed to ``fit`` on the estimator.

        Returns:
            self:
                Return self.
        """

        self._check_params()

        random_state = check_random_state(self.random_state)
        max_samples = self.subsample
        n_samples = _num_samples(X)
        old_level = logger.getEffectiveLevel()

        if self.verbose > 1:
            logger.setLevel(DEBUG)
        elif self.verbose > 0:
            logger.setLevel(INFO)
        else:
            logger.setLevel(WARNING)

        self.sample_indices_ = np.arange(n_samples)

        if type(max_samples) is float:
            max_samples = int(max_samples * n_samples)

        if max_samples < n_samples:
            self.sample_indices_ = random_state.choice(
                self.sample_indices_,
                max_samples,
                replace=False
            )

            self.sample_indices_.sort()

        X_res = safe_indexing(X, self.sample_indices_)
        y_res = safe_indexing(y, self.sample_indices_)
        groups_res = safe_indexing(groups, self.sample_indices_)
        fit_params_res = fit_params

        if fit_params_res is not None:
            fit_params_res = {
                key: _index_param_value(
                    X,
                    value,
                    self.sample_indices_
                ) for key, value in fit_params.items()
            }

        classifier = is_classifier(self.estimator)
        cv = check_cv(self.cv, y_res, classifier)

        self.n_splits_ = cv.get_n_splits(X_res, y_res, groups=groups_res)
        self.scorer_ = check_scoring(self.estimator, scoring=self.scoring)

        if self.study is None:
            seed = random_state.randint(0, np.iinfo('int32').max)
            sampler = samplers.TPESampler(seed=seed)

            self.study_ = study_module.create_study(
                direction='maximize',
                sampler=sampler
            )

        else:
            self.study_ = self.study

        objective = _Objective(
            self.estimator,
            self.param_distributions,
            X_res,
            y_res,
            cv,
            self.enable_pruning,
            self.error_score,
            fit_params_res,
            groups_res,
            self.max_iter,
            self.return_train_score,
            self.scorer_
        )

        logger.info(
            'Searching the best hyperparameters using {} '
            'samples...'.format(_num_samples(self.sample_indices_))
        )

        self.study_.optimize(
            objective,
            n_jobs=self.n_jobs,
            n_trials=self.n_trials,
            timeout=self.timeout
        )

        logger.info('Finished hyperparemeter search!')

        if self.refit:
            self._refit(X, y, **fit_params)

        logger.setLevel(old_level)

        return self

    def score(
        self,
        X,  # type: TwoDimArrayLikeType
        y=None,  # type: Optional[Union[OneDimArrayLikeType, TwoDimArrayLikeType]]
    ):
        # type: (...) -> float
        """Return the score on the given data.

        Args:
            X:
                Data.

            y:
                Target variable.

        Returns:
            score:
                Scaler score.
        """

        return self.scorer_(self.best_estimator_, X, y)



from mock import patch
import pytest

import optuna
from optuna.integration import CmaEsSampler
from optuna.integration import SkoptSampler
from optuna.testing.sampler import FirstTrialOnlyRandomSampler

if optuna.type_checking.TYPE_CHECKING:
    from typing import Callable  # NOQA

    from optuna.samplers import BaseSampler  # NOQA

parametrize_sampler = pytest.mark.parametrize('sampler_class', [
    optuna.integration.SkoptSampler,
    optuna.integration.CmaEsSampler,
])


@pytest.mark.parametrize('sampler_class', [
    lambda: SkoptSampler(
        independent_sampler=FirstTrialOnlyRandomSampler(), skopt_kwargs={'n_initial_points': 5}),
    lambda: CmaEsSampler(independent_sampler=FirstTrialOnlyRandomSampler()),
])
def test_suggested_value(sampler_class):
    # type: (Callable[[], BaseSampler]) -> None

    sampler = sampler_class()
    # direction='minimize'
    study = optuna.create_study(sampler=sampler, direction='minimize')
    study.optimize(_objective, n_trials=10, catch=())
    for trial in study.trials:
        for param_name, param_value in trial.params.items():
            distribution = trial.distributions[param_name]
            param_value_in_internal_repr = distribution.to_internal_repr(param_value)
            assert distribution._contains(param_value_in_internal_repr)

    # direction='maximize'
    study = optuna.create_study(sampler=sampler, direction='maximize')
    study.optimize(_objective, n_trials=10, catch=())
    for trial in study.trials:
        for param_name, param_value in trial.params.items():
            distribution = trial.distributions[param_name]
            param_value_in_internal_repr = distribution.to_internal_repr(param_value)
            assert distribution._contains(param_value_in_internal_repr)


@parametrize_sampler
def test_sample_independent(sampler_class):
    # type: (Callable[[], BaseSampler]) -> None

    sampler = sampler_class()
    study = optuna.create_study(sampler=sampler)

    # First trial.
    def objective0(trial):
        # type: (optuna.trial.Trial) -> float

        p0 = trial.suggest_uniform('p0', 0, 10)
        p1 = trial.suggest_loguniform('p1', 1, 10)
        p2 = trial.suggest_int('p2', 0, 10)
        p3 = trial.suggest_discrete_uniform('p3', 0, 9, 3)
        p4 = trial.suggest_categorical('p4', ['10', '20', '30'])
        return p0 + p1 + p2 + p3 + int(p4)

    with patch.object(sampler, 'sample_independent') as mock_object:
        mock_object.side_effect = [1, 2, 3, 3, '10']

        study.optimize(objective0, n_trials=1)

        # In first trial, all parameters were suggested via `sample_independent`.
        assert mock_object.call_count == 5

    # Second trial.
    def objective1(trial):
        # type: (optuna.trial.Trial) -> float

        # p0, p2 and p4 are deleted.
        p1 = trial.suggest_loguniform('p1', 1, 10)
        p3 = trial.suggest_discrete_uniform('p3', 0, 9, 3)

        # p5 is added.
        p5 = trial.suggest_uniform('p5', 0, 1)

        return p1 + p3 + p5

    with patch.object(sampler, 'sample_independent') as mock_object:
        mock_object.side_effect = [0]

        study.optimize(objective1, n_trials=1)

        assert [call[1][2] for call in mock_object.mock_calls] == ['p5']

    # Third trial.
    def objective2(trial):
        # type: (optuna.trial.Trial) -> float

        p1 = trial.suggest_loguniform('p1', 50, 100)  # The range has been changed
        p3 = trial.suggest_discrete_uniform('p3', 0, 9, 3)
        p5 = trial.suggest_uniform('p5', 0, 1)

        return p1 + p3 + p5

    with patch.object(sampler, 'sample_independent') as mock_object:
        mock_object.side_effect = [90, 0.2]

        study.optimize(objective2, n_trials=1)

        assert [call[1][2] for call in mock_object.mock_calls] == ['p1', 'p5']


@pytest.mark.parametrize('sampler_class', [
    lambda x: SkoptSampler(warn_independent_sampling=x),
    lambda x: CmaEsSampler(warn_independent_sampling=x),
])
def test_warn_independent_sampling(sampler_class):
    # type: (Callable[[bool], BaseSampler]) -> None

    # warn_independent_sampling=True
    sampler = sampler_class(True)
    study = optuna.create_study(sampler=sampler)

    class_name = 'optuna.integration.{}'.format(sampler.__class__.__name__)
    method_name = '{}._log_independent_sampling'.format(class_name)

    with patch(method_name) as mock_object:
        study.optimize(
            lambda t: t.suggest_uniform('p0', 0, 10) + t.suggest_uniform('q0', 0, 10), n_trials=1)
        assert mock_object.call_count == 0

    with patch(method_name) as mock_object:
        study.optimize(
            lambda t: t.suggest_uniform('p1', 0, 10) + t.suggest_uniform('q1', 0, 10), n_trials=1)
        assert mock_object.call_count == 2

    # warn_independent_sampling=False
    sampler = sampler_class(False)
    study = optuna.create_study(sampler=sampler)

    with patch(method_name) as mock_object:
        study.optimize(
            lambda t: t.suggest_uniform('p0', 0, 10) + t.suggest_uniform('q0', 0, 10), n_trials=1)
        assert mock_object.call_count == 0

    with patch(method_name) as mock_object:
        study.optimize(
            lambda t: t.suggest_uniform('p1', 0, 10) + t.suggest_uniform('q1', 0, 10), n_trials=1)
        assert mock_object.call_count == 0


def _objective(trial):
    # type: (optuna.trial.Trial) -> float

    p0 = trial.suggest_uniform('p0', -3.3, 5.2)
    p1 = trial.suggest_uniform('p1', 2.0, 2.0)
    p2 = trial.suggest_loguniform('p2', 0.0001, 0.3)
    p3 = trial.suggest_loguniform('p3', 1.1, 1.1)
    p4 = trial.suggest_int('p4', -100, 8)
    p5 = trial.suggest_int('p5', -20, -20)
    p6 = trial.suggest_discrete_uniform('p6', 10, 20, 2)
    p7 = trial.suggest_discrete_uniform('p7', 0.1, 1.0, 0.1)
    p8 = trial.suggest_discrete_uniform('p8', 2.2, 2.2, 0.5)
    p9 = trial.suggest_categorical('p9', ['9', '3', '0', '8'])

    return p0 + p1 + p2 + p3 + p4 + p5 + p6 + p7 + p8 + int(p9)



import json
import pytest

from optuna import distributions
from optuna import type_checking

if type_checking.TYPE_CHECKING:
    from typing import Any  # NOQA
    from typing import Dict  # NOQA
    from typing import List  # NOQA

EXAMPLE_DISTRIBUTIONS = {
    'u': distributions.UniformDistribution(low=1., high=2.),
    'l': distributions.LogUniformDistribution(low=0.001, high=100),
    'du': distributions.DiscreteUniformDistribution(low=1., high=10., q=2.),
    'iu': distributions.IntUniformDistribution(low=1, high=10),
    'c1': distributions.CategoricalDistribution(choices=(2.71, -float('inf'))),
    'c2': distributions.CategoricalDistribution(choices=('Roppongi', 'Azabu'))
}  # type: Dict[str, Any]

EXAMPLE_JSONS = {
    'u': '{"name": "UniformDistribution", "attributes": {"low": 1.0, "high": 2.0}}',
    'l': '{"name": "LogUniformDistribution", "attributes": {"low": 0.001, "high": 100}}',
    'du': '{"name": "DiscreteUniformDistribution",'
    '"attributes": {"low": 1.0, "high": 10.0, "q": 2.0}}',
    'iu': '{"name": "IntUniformDistribution", "attributes": {"low": 1, "high": 10}}',
    'c1': '{"name": "CategoricalDistribution", "attributes": {"choices": [2.71, -Infinity]}}',
    'c2': '{"name": "CategoricalDistribution", "attributes": {"choices": ["Roppongi", "Azabu"]}}'
}


def test_json_to_distribution():
    # type: () -> None

    for key in EXAMPLE_JSONS.keys():
        distribution_actual = distributions.json_to_distribution(EXAMPLE_JSONS[key])
        assert distribution_actual == EXAMPLE_DISTRIBUTIONS[key]

    unknown_json = '{"name": "UnknownDistribution", "attributes": {"low": 1.0, "high": 2.0}}'
    pytest.raises(ValueError, lambda: distributions.json_to_distribution(unknown_json))


def test_distribution_to_json():
    # type: () -> None

    for key in EXAMPLE_JSONS.keys():
        json_actual = distributions.distribution_to_json(EXAMPLE_DISTRIBUTIONS[key])
        assert json.loads(json_actual) == json.loads(EXAMPLE_JSONS[key])


def test_check_distribution_compatibility():
    # type: () -> None

    # test the same distribution
    for key in EXAMPLE_JSONS.keys():
        distributions.check_distribution_compatibility(EXAMPLE_DISTRIBUTIONS[key],
                                                       EXAMPLE_DISTRIBUTIONS[key])

    # test different distribution classes
    pytest.raises(
        ValueError, lambda: distributions.check_distribution_compatibility(
            EXAMPLE_DISTRIBUTIONS['u'], EXAMPLE_DISTRIBUTIONS['l']))

    # test dynamic value range (CategoricalDistribution)
    pytest.raises(
        ValueError, lambda: distributions.check_distribution_compatibility(
            EXAMPLE_DISTRIBUTIONS['c2'], EXAMPLE_DISTRIBUTIONS['c2']._replace(
                choice=('Roppongi', 'Akasaka'))))

    # test dynamic value range (except CategoricalDistribution)
    distributions.check_distribution_compatibility(
        EXAMPLE_DISTRIBUTIONS['u'], EXAMPLE_DISTRIBUTIONS['u']._replace(low=-1.0, high=-2.0))
    distributions.check_distribution_compatibility(
        EXAMPLE_DISTRIBUTIONS['l'], EXAMPLE_DISTRIBUTIONS['l']._replace(low=-0.1, high=1.0))
    distributions.check_distribution_compatibility(
        EXAMPLE_DISTRIBUTIONS['du'], EXAMPLE_DISTRIBUTIONS['du']._replace(
            low=-1.0, high=10.0, q=3.))
    distributions.check_distribution_compatibility(
        EXAMPLE_DISTRIBUTIONS['iu'], EXAMPLE_DISTRIBUTIONS['iu']._replace(low=-1, high=1))


def test_contains():
    # type: () -> None

    u = distributions.UniformDistribution(low=1., high=2.)
    assert not u._contains(0.9)
    assert u._contains(1)
    assert u._contains(1.5)
    assert not u._contains(2)

    lu = distributions.LogUniformDistribution(low=0.001, high=100)
    assert not lu._contains(0.0)
    assert lu._contains(0.001)
    assert lu._contains(12.3)
    assert not lu._contains(100)

    du = distributions.DiscreteUniformDistribution(low=1., high=10., q=2.)
    assert not du._contains(0.9)
    assert du._contains(1.0)
    assert du._contains(3.5)
    assert du._contains(6)
    assert du._contains(10)
    assert not du._contains(10.1)

    iu = distributions.IntUniformDistribution(low=1, high=10)
    assert not iu._contains(0.9)
    assert iu._contains(1)
    assert iu._contains(3.5)
    assert iu._contains(6)
    assert iu._contains(10)
    assert iu._contains(10.1)
    assert not iu._contains(11)

    c = distributions.CategoricalDistribution(choices=('Roppongi', 'Azabu'))
    assert not c._contains(-1)
    assert c._contains(0)
    assert c._contains(1)
    assert c._contains(1.5)
    assert not c._contains(3)


def test_empty_range_contains():
    # type: () -> None

    u = distributions.UniformDistribution(low=1.0, high=1.0)
    assert not u._contains(0.9)
    assert u._contains(1.0)
    assert not u._contains(1.1)

    lu = distributions.LogUniformDistribution(low=1.0, high=1.0)
    assert not lu._contains(0.9)
    assert lu._contains(1.0)
    assert not lu._contains(1.1)

    du = distributions.DiscreteUniformDistribution(low=1.0, high=1.0, q=2.0)
    assert not du._contains(0.9)
    assert du._contains(1.0)
    assert not du._contains(1.1)

    iu = distributions.IntUniformDistribution(low=1, high=1)
    assert not iu._contains(0)
    assert iu._contains(1)
    assert not iu._contains(2)


def test_single():
    # type: () -> None

    single_distributions = [
        distributions.UniformDistribution(low=1.0, high=1.0),
        distributions.LogUniformDistribution(low=7.3, high=7.3),
        distributions.DiscreteUniformDistribution(low=2.22, high=2.22, q=0.1),
        distributions.IntUniformDistribution(low=-123, high=-123),
        distributions.CategoricalDistribution(choices=('foo', ))
    ]  # type: List[distributions.BaseDistribution]
    for distribution in single_distributions:
        assert distribution.single()

    nonsingle_distributions = [
        distributions.UniformDistribution(low=1.0, high=1.001),
        distributions.LogUniformDistribution(low=7.3, high=10),
        distributions.DiscreteUniformDistribution(low=-30, high=-20, q=2),
        distributions.IntUniformDistribution(low=-123, high=0),
        distributions.CategoricalDistribution(choices=('foo', 'bar'))
    ]  # type: List[distributions.BaseDistribution]
    for distribution in nonsingle_distributions:
        assert not distribution.single()


def test_empty_distribution():
    # type: () -> None

    # Empty distributions cannot be instantiated.
    with pytest.raises(ValueError):
        distributions.UniformDistribution(low=0.0, high=-100.0)

    with pytest.raises(ValueError):
        distributions.LogUniformDistribution(low=7.3, high=7.2),

    with pytest.raises(ValueError):
        distributions.DiscreteUniformDistribution(low=-30, high=-40, q=3),

    with pytest.raises(ValueError):
        distributions.IntUniformDistribution(low=123, high=100),

    with pytest.raises(ValueError):
        distributions.CategoricalDistribution(choices=()),



import copy
from datetime import datetime
import threading

from optuna import distributions  # NOQA
from optuna.storages import base
from optuna.storages.base import DEFAULT_STUDY_NAME_PREFIX
from optuna import structs
from optuna import type_checking

if type_checking.TYPE_CHECKING:
    from typing import Any  # NOQA
    from typing import Dict  # NOQA
    from typing import List  # NOQA
    from typing import Optional  # NOQA

IN_MEMORY_STORAGE_STUDY_ID = 0
IN_MEMORY_STORAGE_STUDY_UUID = '00000000-0000-0000-0000-000000000000'


class InMemoryStorage(base.BaseStorage):
    """Storage class that stores data in memory of the Python process.

    This class is not supposed to be directly accessed by library users.
    """

    def __init__(self):
        # type: () -> None
        self.trials = []  # type: List[structs.FrozenTrial]
        self.param_distribution = {}  # type: Dict[str, distributions.BaseDistribution]
        self.direction = structs.StudyDirection.NOT_SET
        self.study_user_attrs = {}  # type: Dict[str, Any]
        self.study_system_attrs = {}  # type: Dict[str, Any]
        self.study_name = DEFAULT_STUDY_NAME_PREFIX + IN_MEMORY_STORAGE_STUDY_UUID  # type: str
        self.best_trial_id = None  # type: Optional[int]

        self._lock = threading.RLock()

    def __getstate__(self):
        # type: () -> Dict[Any, Any]
        state = self.__dict__.copy()
        del state['_lock']
        return state

    def __setstate__(self, state):
        # type: (Dict[Any, Any]) -> None
        self.__dict__.update(state)
        self._lock = threading.RLock()

    def create_new_study(self, study_name=None):
        # type: (Optional[str]) -> int

        if study_name is not None:
            self.study_name = study_name

        return IN_MEMORY_STORAGE_STUDY_ID  # TODO(akiba)

    def delete_study(self, study_id):
        # type: (int) -> None

        self._check_study_id(study_id)

        with self._lock:
            self.trials = []
            self.best_trial_id = None
            self.param_distribution = {}
            self.direction = structs.StudyDirection.NOT_SET
            self.study_user_attrs = {}
            self.study_system_attrs = {}
            self.study_name = DEFAULT_STUDY_NAME_PREFIX + IN_MEMORY_STORAGE_STUDY_UUID

    def set_study_direction(self, study_id, direction):
        # type: (int, structs.StudyDirection) -> None

        with self._lock:
            if self.direction != structs.StudyDirection.NOT_SET and self.direction != direction:
                raise ValueError('Cannot overwrite study direction from {} to {}.'.format(
                    self.direction, direction))
            self.direction = direction

    def set_study_user_attr(self, study_id, key, value):
        # type: (int, str, Any) -> None

        with self._lock:
            self.study_user_attrs[key] = value

    def set_study_system_attr(self, study_id, key, value):
        # type: (int, str, Any) -> None

        with self._lock:
            self.study_system_attrs[key] = value

    def get_study_id_from_name(self, study_name):
        # type: (str) -> int

        if study_name != self.study_name:
            raise ValueError("No such study {}.".format(study_name))

        return IN_MEMORY_STORAGE_STUDY_ID

    def get_study_id_from_trial_id(self, trial_id):
        # type: (int) -> int

        return IN_MEMORY_STORAGE_STUDY_ID

    def get_study_name_from_id(self, study_id):
        # type: (int) -> str

        self._check_study_id(study_id)
        return self.study_name

    def get_study_direction(self, study_id):
        # type: (int) -> structs.StudyDirection

        return self.direction

    def get_study_user_attrs(self, study_id):
        # type: (int) -> Dict[str, Any]

        with self._lock:
            return copy.deepcopy(self.study_user_attrs)

    def get_study_system_attrs(self, study_id):
        # type: (int) -> Dict[str, Any]

        with self._lock:
            return copy.deepcopy(self.study_system_attrs)

    def get_all_study_summaries(self):
        # type: () -> List[structs.StudySummary]

        best_trial = None
        if any(t for t in self.trials if t.state == structs.TrialState.COMPLETE):
            best_trial = self.get_best_trial(IN_MEMORY_STORAGE_STUDY_ID)

        datetime_start = None
        if len(self.trials) > 0:
            datetime_start = min([t.datetime_start for t in self.trials])

        return [
            structs.StudySummary(
                study_id=IN_MEMORY_STORAGE_STUDY_ID,
                study_name=self.study_name,
                direction=self.direction,
                best_trial=best_trial,
                user_attrs=copy.deepcopy(self.study_user_attrs),
                system_attrs=copy.deepcopy(self.study_system_attrs),
                n_trials=len(self.trials),
                datetime_start=datetime_start)
        ]

    def create_new_trial(self, study_id, template_trial=None):
        # type: (int, Optional[structs.FrozenTrial]) -> int

        self._check_study_id(study_id)

        if template_trial is None:
            trial = self._create_running_trial()
        else:
            trial = copy.deepcopy(template_trial)

        with self._lock:
            trial_id = len(self.trials)
            trial.system_attrs['_number'] = trial_id
            self.trials.append(trial._replace(number=trial_id, trial_id=trial_id))
            self._update_cache(trial_id)
        return trial_id

    @staticmethod
    def _create_running_trial():
        # type: () -> structs.FrozenTrial

        return structs.FrozenTrial(
            trial_id=-1,  # dummy value.
            number=-1,  # dummy value.
            state=structs.TrialState.RUNNING,
            params={},
            distributions={},
            user_attrs={},
            system_attrs={},
            value=None,
            intermediate_values={},
            datetime_start=datetime.now(),
            datetime_complete=None)

    def set_trial_state(self, trial_id, state):
        # type: (int, structs.TrialState) -> None

        with self._lock:
            self.check_trial_is_updatable(trial_id, self.trials[trial_id].state)

            self.trials[trial_id] = self.trials[trial_id]._replace(state=state)
            if state.is_finished():
                self.trials[trial_id] = \
                    self.trials[trial_id]._replace(datetime_complete=datetime.now())
                self._update_cache(trial_id)

    def set_trial_param(self, trial_id, param_name, param_value_internal, distribution):
        # type: (int, str, float, distributions.BaseDistribution) -> bool

        with self._lock:
            self.check_trial_is_updatable(trial_id, self.trials[trial_id].state)

            # Check param distribution compatibility with previous trial(s).
            if param_name in self.param_distribution:
                distributions.check_distribution_compatibility(self.param_distribution[param_name],
                                                               distribution)

            # Check param has not been set; otherwise, return False.
            if param_name in self.trials[trial_id].params:
                return False

            # Set param distribution.
            self.param_distribution[param_name] = distribution

            # Set param.
            self.trials[trial_id].params[param_name] = distribution.to_external_repr(
                param_value_internal)
            self.trials[trial_id].distributions[param_name] = distribution

            return True

    def get_trial_number_from_id(self, trial_id):
        # type: (int) -> int

        return trial_id

    def get_best_trial(self, study_id):
        # type: (int) -> structs.FrozenTrial

        if self.best_trial_id is None:
            raise ValueError('No trials are completed yet.')
        return self.get_trial(self.best_trial_id)

    def get_trial_param(self, trial_id, param_name):
        # type: (int, str) -> float

        distribution = self.trials[trial_id].distributions[param_name]
        return distribution.to_internal_repr(self.trials[trial_id].params[param_name])

    def set_trial_value(self, trial_id, value):
        # type: (int, float) -> None

        with self._lock:
            self.check_trial_is_updatable(trial_id, self.trials[trial_id].state)

            self.trials[trial_id] = self.trials[trial_id]._replace(value=value)

    def _update_cache(self, trial_id):
        # type: (int) -> None

        if self.trials[trial_id].state != structs.TrialState.COMPLETE:
            return

        if self.best_trial_id is None:
            self.best_trial_id = trial_id
            return
        best_value = self.trials[self.best_trial_id].value
        new_value = self.trials[trial_id].value
        if best_value is None:
            self.best_trial_id = trial_id
            return
        # Complete trials do not have `None` values.
        assert new_value is not None

        if (self.get_study_direction(IN_MEMORY_STORAGE_STUDY_ID) ==
                structs.StudyDirection.MAXIMIZE):
            if best_value < new_value:
                self.best_trial_id = trial_id
            return
        if best_value > new_value:
            self.best_trial_id = trial_id
        return

    def set_trial_intermediate_value(self, trial_id, step, intermediate_value):
        # type: (int, int, float) -> bool

        with self._lock:
            self.check_trial_is_updatable(trial_id, self.trials[trial_id].state)

            values = self.trials[trial_id].intermediate_values
            if step in values:
                return False

            values[step] = intermediate_value

            return True

    def set_trial_user_attr(self, trial_id, key, value):
        # type: (int, str, Any) -> None

        with self._lock:
            self.check_trial_is_updatable(trial_id, self.trials[trial_id].state)

            self.trials[trial_id].user_attrs[key] = value

    def set_trial_system_attr(self, trial_id, key, value):
        # type: (int, str, Any) -> None

        with self._lock:
            self.check_trial_is_updatable(trial_id, self.trials[trial_id].state)

            self.trials[trial_id].system_attrs[key] = value

    def get_trial(self, trial_id):
        # type: (int) -> structs.FrozenTrial

        with self._lock:
            return copy.deepcopy(self.trials[trial_id])

    def get_all_trials(self, study_id):
        # type: (int) -> List[structs.FrozenTrial]

        self._check_study_id(study_id)
        with self._lock:
            return copy.deepcopy(self.trials)

    def get_n_trials(self, study_id, state=None):
        # type: (int, Optional[structs.TrialState]) -> int

        self._check_study_id(study_id)
        if state is None:
            return len(self.trials)

        return len([t for t in self.trials if t.state == state])

    def _check_study_id(self, study_id):
        # type: (int) -> None

        if study_id != IN_MEMORY_STORAGE_STUDY_ID:
            raise ValueError('study_id is supposed to be {} in {}.'.format(
                IN_MEMORY_STORAGE_STUDY_ID, self.__class__.__name__))



from __future__ import absolute_import

import numpy as np

import optuna
from optuna import distributions
from optuna import samplers
from optuna.samplers import BaseSampler
from optuna import structs
from optuna.structs import StudyDirection
from optuna import type_checking

try:
    import skopt
    from skopt.space import space

    _available = True
except ImportError as e:
    _import_error = e
    # SkoptSampler is disabled because Scikit-Optimize is not available.
    _available = False

if type_checking.TYPE_CHECKING:
    from typing import Any  # NOQA
    from typing import Dict  # NOQA
    from typing import List  # NOQA
    from typing import Optional  # NOQA
    from typing import Tuple  # NOQA

    from optuna.distributions import BaseDistribution  # NOQA
    from optuna.structs import FrozenTrial  # NOQA
    from optuna.study import Study  # NOQA


class SkoptSampler(BaseSampler):
    """Sampler using Scikit-Optimize as the backend.

    Example:

        Optimize a simple quadratic function by using :class:`~optuna.integration.SkoptSampler`.

        .. code::

                def objective(trial):
                    x = trial.suggest_uniform('x', -10, 10)
                    y = trial.suggest_int('y', 0, 10)
                    return x**2 + y

                sampler = optuna.integration.SkoptSampler()
                study = optuna.create_study(sampler=sampler)
                study.optimize(objective, n_trials=100)

    Args:
        independent_sampler:
            A :class:`~optuna.samplers.BaseSampler` instance that is used for independent
            sampling. The parameters not contained in the relative search space are sampled
            by this sampler.
            The search space for :class:`~optuna.integration.SkoptSampler` is determined by
            :func:`~optuna.samplers.intersection_search_space()`.

            If :obj:`None` is specified, :class:`~optuna.samplers.RandomSampler` is used
            as the default.

            .. seealso::
                :class:`optuna.samplers` module provides built-in independent samplers
                such as :class:`~optuna.samplers.RandomSampler` and
                :class:`~optuna.samplers.TPESampler`.

        warn_independent_sampling:
            If this is :obj:`True`, a warning message is emitted when
            the value of a parameter is sampled by using an independent sampler.

            Note that the parameters of the first trial in a study are always sampled
            via an independent sampler, so no warning messages are emitted in this case.
        skopt_kwargs:
            Keyword arguments passed to the constructor of
            `skopt.Optimizer <https://scikit-optimize.github.io/#skopt.Optimizer>`_
            class.

            Note that ``dimensions`` argument in ``skopt_kwargs`` will be ignored
            because it is added by :class:`~optuna.integration.SkoptSampler` automatically.

    """

    def __init__(self, independent_sampler=None, warn_independent_sampling=True,
                 skopt_kwargs=None):
        # type: (Optional[BaseSampler], bool, Optional[Dict[str, Any]]) -> None

        _check_skopt_availability()

        self._skopt_kwargs = skopt_kwargs or {}
        if 'dimensions' in self._skopt_kwargs:
            del self._skopt_kwargs['dimensions']

        self._independent_sampler = independent_sampler or samplers.RandomSampler()
        self._warn_independent_sampling = warn_independent_sampling

    def infer_relative_search_space(self, study, trial):
        # type: (Study, FrozenTrial) -> Dict[str, BaseDistribution]

        search_space = {}
        for name, distribution in samplers.intersection_search_space(study).items():
            if distribution.single():
                if not isinstance(distribution, distributions.CategoricalDistribution):
                    # `skopt` cannot handle non-categorical distributions that contain just
                    # a single value, so we skip this distribution.
                    #
                    # Note that `Trial` takes care of this distribution during suggestion.
                    continue

            search_space[name] = distribution

        return search_space

    def sample_relative(self, study, trial, search_space):
        # type: (Study, FrozenTrial, Dict[str, BaseDistribution]) -> Dict[str, Any]

        if len(search_space) == 0:
            return {}

        optimizer = _Optimizer(search_space, self._skopt_kwargs)
        optimizer.tell(study)
        return optimizer.ask()

    def sample_independent(self, study, trial, param_name, param_distribution):
        # type: (Study, FrozenTrial, str, BaseDistribution) -> Any

        if self._warn_independent_sampling:
            complete_trials = [t for t in study.trials if t.state == structs.TrialState.COMPLETE]
            if len(complete_trials) >= 1:
                self._log_independent_sampling(trial, param_name)

        return self._independent_sampler.sample_independent(study, trial, param_name,
                                                            param_distribution)

    def _log_independent_sampling(self, trial, param_name):
        # type: (FrozenTrial, str) -> None

        logger = optuna.logging.get_logger(__name__)
        logger.warning("The parameter '{}' in trial#{} is sampled independently "
                       "by using `{}` instead of `SkoptSampler` "
                       "(optimization performance may be degraded). "
                       "You can suppress this warning by setting `warn_independent_sampling` "
                       "to `False` in the constructor of `SkoptSampler`, "
                       "if this independent sampling is intended behavior.".format(
                           param_name, trial.number, self._independent_sampler.__class__.__name__))


class _Optimizer(object):
    def __init__(self, search_space, skopt_kwargs):
        # type: (Dict[str, BaseDistribution], Dict[str, Any]) -> None

        self._search_space = search_space

        dimensions = []
        for name, distribution in sorted(self._search_space.items()):
            if isinstance(distribution, distributions.UniformDistribution):
                # Convert the upper bound from exclusive (optuna) to inclusive (skopt).
                high = np.nextafter(distribution.high, float('-inf'))
                dimension = space.Real(distribution.low, high)
            elif isinstance(distribution, distributions.LogUniformDistribution):
                # Convert the upper bound from exclusive (optuna) to inclusive (skopt).
                high = np.nextafter(distribution.high, float('-inf'))
                dimension = space.Real(distribution.low, high, prior='log-uniform')
            elif isinstance(distribution, distributions.IntUniformDistribution):
                dimension = space.Integer(distribution.low, distribution.high)
            elif isinstance(distribution, distributions.DiscreteUniformDistribution):
                count = (distribution.high - distribution.low) // distribution.q
                dimension = space.Integer(0, count)
            elif isinstance(distribution, distributions.CategoricalDistribution):
                dimension = space.Categorical(distribution.choices)
            else:
                raise NotImplementedError(
                    "The distribution {} is not implemented.".format(distribution))

            dimensions.append(dimension)

        self._optimizer = skopt.Optimizer(dimensions, **skopt_kwargs)

    def tell(self, study):
        # type: (Study) -> None

        xs = []
        ys = []
        for trial in study.trials:
            if trial.state != structs.TrialState.COMPLETE:
                continue

            if not self._is_compatible(trial):
                continue

            x, y = self._complete_trial_to_skopt_observation(study, trial)
            xs.append(x)
            ys.append(y)

        self._optimizer.tell(xs, ys)

    def ask(self):
        # type: () -> Dict[str, Any]

        params = {}
        param_values = self._optimizer.ask()
        for (name, distribution), value in zip(sorted(self._search_space.items()), param_values):
            if isinstance(distribution, distributions.DiscreteUniformDistribution):
                value = value * distribution.q + distribution.low

            params[name] = value

        return params

    def _is_compatible(self, trial):
        # type: (FrozenTrial) -> bool

        # Thanks to `intersection_search_space()` function, in sequential optimization,
        # the parameters of complete trials are always compatible with the search space.
        #
        # However, in distributed optimization, incompatible trials may complete on a worker
        # just after a product search space is calculated on another worker.

        for name, distribution in self._search_space.items():
            if name not in trial.params:
                return False

            distributions.check_distribution_compatibility(distribution, trial.distributions[name])
            param_value = trial.params[name]
            param_internal_value = distribution.to_internal_repr(param_value)
            if not distribution._contains(param_internal_value):
                return False

        return True

    def _complete_trial_to_skopt_observation(self, study, trial):
        # type: (Study, FrozenTrial) -> Tuple[List[Any], float]

        param_values = []
        for name, distribution in sorted(self._search_space.items()):
            param_value = trial.params[name]

            if isinstance(distribution, distributions.DiscreteUniformDistribution):
                param_value = (param_value - distribution.low) // distribution.q

            param_values.append(param_value)

        value = trial.value
        assert value is not None

        if study.direction == StudyDirection.MAXIMIZE:
            value = -value

        return param_values, value


def _check_skopt_availability():
    # type: () -> None

    if not _available:
        raise ImportError(
            'Scikit-Optimize is not available. Please install it to use this feature. '
            'Scikit-Optimize can be installed by executing `$ pip install scikit-optimize`. '
            'For further information, please refer to the installation guide of Scikit-Optimize. '
            '(The actual import error is as follows: ' + str(_import_error) + ')')
